// =============================================================================
//
// Introduction to Dafny, the verification language
//
// =============================================================================
//
// At its core Dafny is an imperative programming language. Dafny also
// supports programme verification using the /Design-by-Contract/
// methodology. For verification, Dafny adopts a functional approach
// to writing specifications.
//
// ## What we will cover
//
// This file demonstrates, some of Dafny's verification abilities.
// Specifically, we will explore how Dafny approaches Design-by-Contract and
// use this to reason about simple programs. Looking at:
//
// 0. Functional and Imperative Programming;
// 1. Compile-time verification using Pre-and-Post Conditions and Invariants;
// 2. Runtime verification using tests, asserts, and expects;
// 3. Predicates for making readable specifications;
// 4. Writing richer specifications for verification;
//
// We will do some coding together, but we will give you this file
// unmodified to try outside of the lecture.
//
// A lengthier look at Dafny for programming is on MyPlace.
//
// ## What we will not cover.
//
// Dafny is a very rich and expressive language in which you *can* do more
// advanced verification, and programming. For instance, Dafny supports lemmas,
// co-data, objects,
//
// For CS886, we will not go much beyond these core concepts.
// In future sessions we will look at how Dafny can reason about states and
// heap values, as well as ensuring that some code is compile time only (Ghosts).
//
// ## Thanks to
//
// We base this session off of the
// Dafny Getting Started tutorial.
//
// https://dafny.org/latest/OnlineTutorial/guide


// =============================================================================
//
//  Dafny does Design-by-Contract
//
// =============================================================================


// ## The First Example
//

/** Return the absolute value of an integer.

    Given an integer, \(x\), the **absolute**  value is calculated as:

   \[
    abs(x) = x  if x >= 0;
             -x if x <  0;
   \]
 */
method MyAbs(x : int) returns (y : int)
{
  if x >= 0 {
    return x;
  }
}

// [ Question ]
//
// Is this method correct?

// [ Notes ]
//
// + Here `MyAbs` is a *method* that takes, as arguments, a single integer `x`,
//   and will return an integer.
//
// + Method return values are bound to an identifier as we may need to
//   refer to them in specifications.
//
//

/* [ Marker ] ==================================================================

// =============================================================================

// Part 1.2: Verifying Absolute, Incorrectly

// =============================================================================

// We will now start to write specifications for Absolute.
//
method MyAbsV(x : int)
     returns (y : int)
     ensures y >= 0
{
  return 0;
}

// =============================================================================
//
// Part 1.3: Verifying Absolute Correctly
//
// =============================================================================
//
// Testing programs enable us to validate our specifications. Tests, however,
// are not always complete. Neither are specifications.
//
// So let us be a bit more principled in writing a better specification:
//
// + What is a good set of properties?
// + How does input and output relate?
//
//
// Given an integer, \(x\), the **absolute**  value is calculated as:
//
// \[
//  abs(x) = x  if x >= 0;
//           -x if x <  0;
// \]
//
//

method MyAbsV1(x : int) returns (y : int)
  ensures y >= 0 // output is +ve

  // What else can we say based on the inputs?
{
  return 0;
}

method MyAbsV1(x : int) returns (y : int)
  ensures y >= 0 // output is +ve

  // What else can we say based on the inputs?
{
  return 0;
}

// [ Notes ]
//
// + Specifications can also depend on implications (`==>`) which states:
//
//   If the LHS is true then so must the RHS.
//
// + Your first attempt at writing a specification may result in one that is
//   verbose. Specifications can be simplified.
//

// =============================================================================
//
// Part 2: Predicates, Pre-conditions, and datatypes. Oh My!
//
// =============================================================================
//
// So far we have performed some basic design-by-contract. Let us look
// a more detailed example.
//
// We start with by looking at how to write specifications more tidily
// using predicates.
//
//

predicate isPos(i : int)
{
  i >= 0
}

predicate isNeg(i : int)
{
  i < 0
}

predicate areSame(i : int, j : int)
{
  i == j
}

// [ Notes ]
//
// + Predicates are boolean expressions that must evaluate to `true`.
//
// + Convention is that they are written in lower case.
//
//

// For example:
method MyAbsV2(x : int)
  returns (y : int)
  ensures isPos(y)
  ensures isNeg(x) ==> areSame(y,-x)
  ensures isPos(x) ==> areSame(x,y)

// [ Notes ]
//
// Methods without bodies are prototypes.
//
//

// =============================================================================
//
// Part 2.2: Indexing Arrays
//
// =============================================================================


// Here is a method to index an array that fails to verify.
method MyFind<T> (a : array<T>, idx : int)
  returns (elem : T)
{
  return a[idx]; // <- Fails here
}

// [ Notes ]
//
// + We have parameterised `MyFind` with `<T>`, which states that `MyFind` is
//   polymorphic and works on all arrays. The notation comes from Java/C#;
//
// + `MyFind` fails to verify because we do not know if the index is within
//   range;
//
// + We can make `MyFind` safe in two ways:
//   + Using a safer design;
//   + Using a precise contract;
//

// Let us write a more verified version.
method MyFindV<T> (a : array<T>, idx : int)
  returns (elem : T)
  // What do we need to make the body true?
  // Is the resulting specification complete?
{
  return a[idx];
}


// =============================================================================
//
// Part 3: Functional Specifications; Imperative Code
//
// =============================================================================
//
// We will now look at a slightly more complex specifications and
// implementations.
//
// We can use functions as code, but more importantly they are there
// to write richer specifications!


// Consider this implementation of multiplication for nats:
//
// `nat` are positive integers
//
method Times(x : nat, y : nat)
  returns (z : nat)
{
  z := 0;
  for i := 0 to x {
    z := z + y;
  }
}

// Let us try to verify it:

method TimesV(x : nat, y : nat)
  returns (z : nat)
  ensures z == x * y
{
  z := 0;
  for i := 0 to x {
    z := z + y;
  }
}

// [ Notes ]
//
// + The post condition cannot be proven true because the implementation
//   cannot be inspected.
//
// + Although, we could rewrite the implementation to mirror the specification,
//   not all scenarios will support this.
//
// + We need to weave code and specification.
//

/** A tail recursive specification for multiplication
 */
function timesS(x : nat, y : nat) : nat
  decreases x,y // <- tells us which parameter is decreasing.
{
  if x == 0
    then 0
    else y + timesS(x - 1, y)
}

method TimesV2 (x : nat, y : nat)
  returns (z : nat)
  ensures z == timesS(x,y)
{
  z := 0;
  for i := 0 to x
    invariant z == timesS(i,y)
  {
    z := z + y;
  }
}

// [ Notes ]
//
// + The idea is to write specifications that are _easier_ to reason about,
//   and link with more efficient implementations.
//
// + We introduce the invariant on the for-loop to inform the compiler further
//   over how the code and specifications are linked.
//


// -- EOF


/* [ Marker ] ==================================================================
== [ Marker ] ================================================================ */


// -- [ EOF ]

// =============================================================================
//
// Header
//
// ========================================================================== */

// [ NOTE ]
//
//
