||| =============================================================================
|||
||| Introduction to Idris, the programming language
|||
||| =============================================================================
|||
||| At its core Idris is a functional programming language, but Idris
||| is built on QTT which enables an expressive environment in which
||| we can reason about programs.
|||
||| ## What we will cover
|||
||| This file demonstrates some elementary aspects of Idris as a
||| language for verification.  Specifically, we will explore
||| dependently-typed functional Programming by example.  Looking at:
|||
||| 1. Proposition as Types;
||| 2. Proofs as Programms;
||| 3. Dependent-pattern matching
|||
||| We will do so by aiming for these one objectives:
|||
||| 1. How to add two natural numbers.
|||
||| A lengthier look at Idris for programming is on MyPlace.
|||
||| We will do some coding together, but we will give you this file
||| unmodified to try outside of the lecture.
|||
||| A lengthier look at Idris for programming is on MyPlace.
|||
||| ## What we will not cover.
|||
||| Dependently-typed programming can be very rich and expressive
||| language in which you *can* do more advanced verification, and
||| programming.
|||
||| For CS886, we will not go much beyond the concepts demonstrated
||| here.  In future sessions we will look at how we can introduce
||| more information into types and use that information for
||| reasoning. We will also look at how QTT enables precise reasoning
||| about when and where variables are used.
|||
module Idris

-- =============================================================================
--
-- Representing (Positive) Numbers
--
-- =============================================================================
--
-- Peano numbers are a simple model to reason about arithmetic of
-- Natural/Whole numbers.

data Peano = Zero      -- There is nothing, the base case;
                       -- or
           | Inc Peano -- there is one more.


-- [ NOTE ]
--
-- Within Idris `Peano` is Nat, we will use `Nat` from now on.

namespace Hidden
  data Nat = Z | S Hidden.Nat

  %hide Hidden.Nat

testPlus : 1 + (S Z) = 2

{- [ Marker ] ==================================================================


-- =============================================================================
--
-- Adding Numbers Disfunctionally
--
-- =============================================================================

plus : (a,b : Nat) -> Nat

-- [ Question ]
--
-- Is the definition of `plus` correct?
--
-- Use the REPL to find out!


-- =============================================================================
--
-- Adding Numbers Disfunctionally
--
-- =============================================================================

plus' : (a,b : Nat) -> Nat
plus' Z b
  = b

plus' (S x) b
  = S $ plus' x b

-- [ Question ]
--
-- We know this definition is correct, but how can we:
--
-- 1. Verify that it is correct?
-- 2. Have a practical implementation?

-- =============================================================================
--
-- Correct-by-Construction Implementation
--
-- =============================================================================
--
--
-- Let us begin by encoding the _correct_ rules for addition.
--
-- We will then provide evidence, the proof, that `Plus` *is* correct.
-- That is we first describe our proposition (theory) about addition.
-- Second we will then prove some things about `Plus`. Namely that,
-- addition works and that we can check the result of an addition.
--
-- First we will define what addition *is*!

|||
||| `Plus` is a *type constructor*, a function that returns a `Type`, and takes as arguments:
|||
||| @a A Nat representing the left operand.
||| @b A Nat representing the right operand.
||| @c A Nat representing the result of the addition.
|||
data Plus : (a,b,c : Nat) -> Type
  where
    ||| Base case.
    |||
    ||| ```
    |||    0 + b = b
    ||| ```
    ZI : Plus Z b b

    ||| Induction on left operand.
    |||
    ||| ```
    ||| S(a) + b = S(a+b)
    ||| ```
    IA : (step : Plus     x  b    c)
              -> Plus (S  x) b (S c)

-- [ Note]
--
-- Both `ZI` and `IA` are *data constructors*, functions that
-- construct instances of a type.
--
--

-- =============================================================================
--
-- Existential Quantification
--
-- =============================================================================
--
-- To look at proving things abotu `How do we evidence that `Plus` is correct?
--
-- To do so we need a mechanism, a datatype, to provide 'evidence'
-- that a value inhabits a type constructor.
--
-- That is, for `Plus` we want to state the following:
--
-- #+begin_example
-- Given two natural numbers: a, b.
-- There exists a number c such that the following holds:
--
-- ```
--    a + b = c
-- ```
--
-- We do so using a dependent pair.

||| Dependent pair, witness that their is a value that will inhabit a
||| type.
data DPair : (type : Type)
          -> (p    : type -> Type)
                  -> Type
  where
    MkDPair : (val : type)
           -> (prf : p val)
                  -> DPair type p

%hide Idris.DPair

||| 'DPair', already exists in the Idris standard libraries, we give
||| the actual definition below.
|||
||| The datatype `DPair` is so widely used that there is special
||| syntax that is often used.
|||
||| ```
||| DPair Nat (\c => Plus a b c)
||| ```
|||
||| is the same as:
|||
||| ```
||| (x : Nat ** Plus a b c)
||| ```
namespace Hidden
  record DPair type (p : type -> Type) where
    constructor MkDPair
    fst : type
    snd : p fst

  -- [ NOTE ]
  --
  -- Records are special datatypes that turn into regular (dependent)
  -- datatypes, and the fields get turned into functions.

  %hide Hidden.DPair

-- =============================================================================
--
-- Proving addition works
--
-- =============================================================================

plusWorks : (x,y : Nat) -> (c ** Plus x y c)
plusWorks 0 y
  = (y ** ZI)
plusWorks (S k) y with (plusWorks k y)
  plusWorks (S k) y | (fst ** snd)
    = (S fst ** IA snd)

||| [ Remark ]
|||
||| 'plusWorks` is the same as the following
plusWorks' : (x,y : Nat) -> DPair Nat (Plus x y)
plusWorks' x y = plusWorks x y

-- =============================================================================
--
-- Checking addition works
--
-- =============================================================================
--
-- We also want to check that addition works.
--
-- However this requires that it works *for all inputs*.
--
-- We know that this is not true: \(1 + 1 == 3\)
--
-- The function type signature will type check as it is 'correct' but
-- we won't be able to provide a body that satisfies the signature.
--
plusCheck' : (a,b,c : Nat) -> Plus a b c


|||  [ Intermezzo ] `Decidable`
|||
||| *Decidability* is the property that encodes whether a given proof holds or not.
|||
||| In languages with dependent types we can encode this as `Decidable`.
|||
||| Which is defined in the idris standard libraries are:
|||
namespace Hidden

  data Dec : (pred : type) -> Type where
    Yes : (prf : pred) -> Dec pred
    No  : (prf : pred -> Void) -> Dec pred

  %hide Hidden.Dec

||| A decision procedure for `Plus`.
|||
||| Given three natural numbers, a,b,c, `plusCheck` will return:
||| + `Yes` if \(a + b = c\);
||| + `No` if there is evidence otherwise.
plusCheck : (a,b,c : Nat) -> Dec (Plus a b c)

Uninhabited (Plus 0 0 (S k)) where
  uninhabited ZI impossible
  uninhabited (IA step) impossible

Uninhabited (Plus 0 (S k) 0) where
  uninhabited ZI impossible
  uninhabited (IA step) impossible

Uninhabited (Plus (S k) 0 0) where
  uninhabited ZI impossible
  uninhabited (IA step) impossible

Uninhabited (Plus (S k) (S j) 0) where
  uninhabited ZI impossible
  uninhabited (IA step) impossible

plusCheck 0 0 0
  = Yes ZI

plusCheck 0 0 (S k)
  = No absurd

plusCheck 0 (S k) 0
  = No absurd

plusCheck 0 (S k) (S j) with (plusCheck 0 k j)
  plusCheck 0 (S k) (S k) | (Yes ZI)
    = Yes ZI
  plusCheck 0 (S k) (S j) | (No contra)
    = No (\case ZI => contra ZI)

plusCheck (S k) 0 0
  = No absurd

plusCheck (S k) 0 (S j) with (plusCheck k 0 j)
  plusCheck (S k) 0 (S j) | (Yes prf)
    = Yes (IA prf)
  plusCheck (S k) 0 (S j) | (No contra)
    = No (\case (IA step) => contra step)

plusCheck (S k) (S j) 0
  = No absurd

plusCheck (S k) (S j) (S i) with (plusCheck k (S j) i)
  plusCheck (S k) (S j) (S i) | (Yes prf)
    = Yes (IA prf)

  plusCheck (S k) (S j) (S i) | (No contra)
    = No (\case (IA step) => contra step)


||| [ Intermezzo ] Equality
|||
||| We have used the Equality type `=`.
|||
||| Equality, as a type, is defined as follows.
namespace Hidden

  data Equal : (a,b : type) -> Type
    where
      Refl : Equal a a

  -- [ Note ]
  --
  -- It is also bound to an operator `=`.
  %hide Hidden.Equal

-- =============================================================================
--
-- Linking Verified Implementations to Non-Verified ones.
--
-- =============================================================================
--
-- We can use `Equals` to make some lemmas about how things work.

||| Equality is transitive
transitive : {a,b,c : Nat} -- Bound the implicits to show LHS magic.
          -> Equal a b
          -> Equal   b c
          -> Equal a   c


||| We can link proofs to real implementations.
plusSame : (prf : Plus  a b   c)
               -> plus' a b = c


-- =============================================================================
--
-- An Addendum: Decidable Equality of Values
--
-- =============================================================================
--
-- We can use `Decidable` and `Equals` to have better definitions of equality.
--
--
-- First let us look at equality of Nats, before we can see how
-- dependent types provides something stronger.
--
-- The library `Decidable.Equality` provides an interface for describing if two things are equal.
namespace Nat

  ||| Standard definition of comparing to datatypes for equality.
  eq : (x,y : Nat) -> Bool
  eq 0 0
    = True

  eq 0 (S k)
    = False

  eq (S k) 0
    = False
  eq (S k) (S j)
    = eq k j


  Uninhabited (Prelude.Z = S x) where
    uninhabited Refl impossible

  negEqSym : (x = y -> Void) -> y = x -> Void
  negEqSym f Refl = f Refl

  decEq : (x,y : Nat) -> Dec (x = y)
  decEq Z Z = Yes Refl
  decEq Z (S x)
    = No absurd
  decEq (S x) Z
    = No (negEqSym absurd)
  decEq (S x) (S y) with (decEq x y)
    decEq (S x) (S y) | (Yes prf) with (prf)
      decEq (S y) (S y) | (Yes Refl) | Refl
      = Yes Refl
    decEq (S x) (S y) | (No contra)
      = No (\case Refl => contra Refl)




== [ Marker ] =============================================================== -}

-- [ EOF ]

-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--
