{- CS886 Week 4 -- Two Ways to do Bounds Checking using Idris

   In this part of the Live Coding session we will cover approaches to
   bounds checking in Idris2

-}
module Bounds

import Decidable.Equality -- required to prove that two values are equal.


{- =============================================================================

   A Reminder on Cons Lists in Idris2

```
data List a = Nil | (::) a (List a)
```

or

```
data List : (a : Type) -> Type where
  Nil : List a
  (::) : (head : a)
      -> (tail : List a)
              -> List a
```

which really is:

```
data List : (a : Type) -> Type where
  Nil : {0 a : Type} -> List a
  (::) : {0 a : Type}
     -> (head : a)
     -> (tail : List a)
             -> List a
```

 [ NOTE ] Any list-like structure that has `Nil` and `(::)` as data constructor names gains access to list notation.

-}

{- =============================================================================

   Quantification over Lists

-}
||| Namespaces are lightweight modules.
|||
||| by default everything is hidden
namespace List

  namespace NonEmpty

    data NonEmpty : (xs : List a) -> Type where
      IsNonEmpty : NonEmpty (x::xs)

    Uninhabited (NonEmpty Nil) where
      uninhabited IsNonEmpty impossible

    nonEmpty : (xs : List a) -> Dec (NonEmpty xs)
    nonEmpty []
      = No absurd

    nonEmpty (x :: xs)
      = Yes IsNonEmpty


    head : (xs  : List a)
        -> (prf : NonEmpty xs)
        => a
    head (x :: xs) {prf = IsNonEmpty} -- curly brackets as is it a bound implicit
      = x

    tail : (xs : List a)
        -> (prf : NonEmpty xs)
        => List a
    tail (x :: xs) {prf = IsNonEmpty}
      = xs

  namespace Elem

    ||| Existential quantification that `x` is in `xs`.
    |||
    public export -- export both type constructor and data constructors
    data Elem : (x : a) -> (xs : List a) -> Type where
      ||| We have found the list elem
      Here : (prf : x = y)
                -> Elem x (y::xs)

      ||| Not here
      There : (prf   : x = y -> Void)
           -> (later : Elem x     xs)
                    -> Elem x (y::xs)


    Uninhabited (Elem x Nil) where
      uninhabited (Here prf) impossible
      uninhabited (There prf later) impossible

    ||| Decision procedure to check.
    export
    elem : DecEq a                -- Need to compare values for equality
        => (x  : a)               -- `=>` is a typeclass constraint
        -> (xs : List a)
              -> Dec (Elem x xs)






  {-
    elem x []
      = No absurd

    elem x (y :: xs) with (decEq x y)
      elem x (y :: xs) | (Yes prf) with (prf)
        elem x (x :: xs) | (Yes prf) | Refl
          = Yes (Here prf)
      elem x (y :: xs) | (No contra) with (elem x xs)
        elem x (y :: xs) | (No contra) | (Yes prf)
          = Yes (There contra prf)
        elem x (y :: xs) | (No contra) | (No f)
          = No (\case (Here prf) => contra prf
                      (There prf later) => f later)
  -}





    index : (xs  : List a)
         -> (idx : Elem x xs)
                -> a







  --  index (x :: xs) (Here Refl)
  --    = x
  --  index (y :: xs) (There prf later) = index xs later

    ||| Ordering of paramters maters!
    index' : (idx : Elem x xs)
          -> (xs  : List a)
                -> a








  --  index' (Here Refl) [] = ?index'_rhs_3
  --  index' (Here Refl) (y :: xs)
  --     = y
  --  index' (There prf later) xs = ?index'_rhs_1


  ||| Elem is a special case of `Any`,
  namespace Any

    data Any : (p  : (x : a) -> Type) -- Proof about a value `x`.
            -> (xs : List a)          -- A list `xs` where an instance of `p` should exist.
                  -> Type
      where
        ||| Here is where we found the first instance of `p`.
        Here  : {0 p : (x : a) -> Type}
             -> (prf : p x) -> Any p (x::xs)

        ||| It was found later
        There : {0 p  : (x : a) -> Type}
             -> (head : p x -> Void)
             -> (tail : Any p xs)
                     -> Any p (x::xs)


    Uninhabited (Any p []) where
      uninhabited (There prf) impossible

    any : (f  : (x : a) -> Dec (p x))
       -> (xs : List a)
             -> Dec (Any p xs)
    any f []
       = No ?as

    any f (x :: xs) with (f x)
      any f (x :: xs) | (Yes prf)
        = Yes (Here prf)
      any f (x :: xs) | (No contra) with (any f xs)
        any f (x :: xs) | (No contra) | (Yes prf)
          = Yes (There contra prf)
        any f (x :: xs) | (No contra) | (No g)
          = No (\case (Here prf) => contra prf
                      (There head tail) => g tail)


    Elem : DecEq a => (x : a) -> (xs : List a) -> Type
    Elem x xs = Any (Equal x) xs


    elem : DecEq a => (x : a) -> (xs : List a) -> Dec (Any.Elem x xs)
    elem x xs = any (decEq x) xs


  namespace All
    data All : (p  : (x : a) -> Type) -- Proof about a value `x`.
            -> (xs : List a)          -- A list `xs` where all values satisfy `p`
                  -> Type
      where
        ||| Here is where we found the first instance of `p`.
        Nil  : All p Nil

        ||| It was found later
        (::) : {0 p  : (x : a) -> Type}
            -> (head : p x)
            -> (tail : All p xs)
                    -> All p (x::xs)


    all : (f  : (x : a) -> Dec (p x))
       -> (xs : List a)
             -> Dec (All p xs)
    all f []
      = Yes Nil
    all f (x :: xs) with (f x)
      all f (x :: xs) | (Yes prf) with (all f xs)
        all f (x :: xs) | (Yes prf) | (Yes ltr)
          = Yes (prf :: ltr)
        all f (x :: xs) | (Yes prf) | (No contra)
        = No (\case (head :: tail) => contra tail)

      all f (x :: xs) | (No contra)
        = No (\case (head :: tail) => contra head)

{- [ Note ] `Any`, `All`, and `Elem` for lists, and vectors are found
            in Idris2 standard library.
-}




{- =============================================================================

   Lists and Vectors

-}

namespace Vect

  data Vect : (l : Nat) -> (a : Type) -> Type where
    Nil : Vect Z a
    (::) : (head : a)
        -> (tail : Vect    n  a)
                -> Vect (S n) a


  head : Vect (S n) a -> a
  head (x :: xs) = x

  tail : Vect (S n) a -> Vect n a
  tail (x :: xs) = xs

  map : (f : a -> b) -> Vect n a -> Vect n b
  map f [] = []
  map f (x :: xs) = f x :: map f xs

  ||| For indexing we need a helper proposition
  namespace Fin

    ||| Fin is a number that is within the bound (0,n).
    |||
    ||| It is built in so we need to disambiguate
    data Fin : (n : Nat) -> Type where
      FZ : Fin (S n)
      FS : (ltr : Fin n)
               -> Fin (S n)

    index : (idx : Fin n) -> (xs : Vect n a) -> a
    index FZ (head :: tail) = head
    index (FS ltr) (x :: tail) = index ltr tail


{- =============================================================================


-}

-- [ Marker ] ==================================================================



-- [ Marker ] ==================================================================



-- [ EOF ]
