// =============================================================================
//
// Reasoning about Data Hygiene using Design-by-Contract
//
// =============================================================================
//
// TYPESTATE associates types with a state, and the set of operations
// acting over the state follows a state machine.
//
// This file demonstrates using TYPESTATE to reasoning about data hygiene.
//
// We will answer some questions together, but we will give you this
// file unmodified to try outside of the lecture.
//
// Remember: Lengthier, slower, notes about Dafny are on MyPlace.
//

// =============================================================================
//
// Capturing Data Hygiene
//
// =============================================================================

/**
   Represents hygiene of data coming from the outside world.
 */
datatype Hygiene = CLEAN | DIRTY

  // [ NOTE ]
  //
  // Hygiene is an 'enumerated' type that only has data constructors.

  /**
     We will need to compare instances of `Hygiene` together.
   */
  predicate hygieneEq1(a : Hygiene, b : Hygiene)
  {
    match a
    {
    case CLEAN => match b { case CLEAN => true case _ => false}
    case DIRTY => match b { case DIRTY => true case _ => false}
    }
  }

  // [ Question ]
  //
  // Can you write a more readable `hygieneEq1` using predicates for
  // `isClean1

  // [ NOTE ]
  //
  // As `Hygiene` is a datatype, a functional programmer would reach for
  // pattern matching to access the potential values that `Hygiene` can
  // have. But there is another way.
  //


predicate hygieneEq(a : Hygiene, b : Hygiene)
{
  a == b
}

// [ NOTE ]
//
// Dafny supports value equality tests on datatypes, as
// follows. Whilst this is fine for enumerated types, you may wish to
// be careful for more expressive datatypes.
//
// Sometimes default operations are not 'sane'.

/* [ Marker ] ==================================================================

// =============================================================================
//
// Labelling Data
//
// =============================================================================
//
// We will now build labelled data that associates a type with a state.
//
// Here is our first attempt.

/*
   Datum pairs strings with a hygiene status.
 */
datatype Datum1 = D1(value : string, status : Hygiene)

// [ NOTE ]
//
// We have chosen a datatype here, but equally you can use pairs.
//
// We can even make `Datum` generic, but for this week we will.

/*
   Is a given instance of `Datum` clean?
 */
predicate isClean1(r : Datum1)
{
  r.status == CLEAN // <- '.' is an accessor
}

// [ NOTE ]
//
// Rather than have to deconstruct instances of `Datum`, Dafny enables
// one to 'project' the contents of a datatype. Using the '.' symbol.

// [ Question ]
//
// Do we want `Hygiene` to exist at runtime?


// =============================================================================
//
// Ghost Data
//
// =============================================================================
//
//
// Ghost data is data that is *only* used for specification, i.e. it
// only exists at runtime.
//
// When ever we have ghost data we need to ensure that data exists in
// the 'ghost realm'.
//
// We will see examples of this through out the remainder of the file.

datatype Datum = D(value : string, ghost status : Hygiene)

/*
   Is a given instance of `Datum` clean?
 */
ghost predicate isClean(r : Datum)
{
  r.status == CLEAN // <- '.' is an accessor
}


// [ Question ]
//
// Can you write a predicate `IsDirty`?
//
// Can you do it using `Match`?





// =============================================================================
//
// A State Machine for Data Hygenie
//
// =============================================================================
//
// We will now build labelled data that associates a type with a
// state, and maintains that state.
//
// Our state machine is simple:
//
// string -[creation]-> DIRTY -[scrub]-> CLEAN -> string
//
// That is, when raw strings enter `Data` they are labelled dirty.
// Data remains dirty until it is cleaned. Finally we can only access
// the underlying string it it has been cleaned,


/* A factory for labelling data dirty.

   We can use `RawInput` to place strings into a labelled datatype.
 */
method RawInput(s : string) returns (res : Datum)
  ensures !isClean(res) // This ensures the `Data` us *dirty*!
{
  return D(s,DIRTY);
}

// [ NOTE ]
//
// Methods live in an impure world, and contain statements.
//
// There will be times when a more pure (and functional) approach is
// more appropriate.
//
// How and where you need to use methods over functions is context
// dependent.
//
// I tend to remain pure, until i need to be impure. Other ways are
// okay.


// [ Question ]
//
// Can you complete the functional definition of `rawInput`?

function rawInput(s : string) : Datum
  ensures !isClean(rawInput(s))

// [ Question ]
//
// Can you complete the definition of 'scrub', complete with
// specification?
//
// We will assume that cleaning is the following magic identity function:

function magic(s : string) : string { s }

/* Scrub takes, as input, dirty data and returns clean data.
 */
function scrub(s : Datum) : Datum

// [ Question ]
//
// Do we need the pre-condition of data?

// [ NOTE ]
//
// Matching to do a datatype update is fine, but can be very verbose
// in Dafny's imperative syntax.
//
// The `s.( status := CLEAN)` field update operator reduces
// unnecessary syntax.


/* Print extracts cleaned data and displays it to the console.
 */
method PrintLn(d : Datum)
  requires isClean(d)
{
  print d.value, "\n";
}

method {:test} testStateMachine()
{
  var s := "Hello World";

  var d := rawInput(s);
//  expect!isClean(d);

  var c := scrub(d);
//  expect isClean(c);

  // PrintLn(d); // <- remember to uncomment me and see what happens!
  PrintLn(c);
}

// =============================================================================
//
// Collecting Hygiene
//
// =============================================================================
//
// `Datum` is okay, but collecting data is also important.

/**
   A collection of Datum, Data.
 */
type Data = seq<Datum>

// [ Question ]
//
// What is 'wrong' with this definition?
//



type DirtyData
  = xs : seq<Datum>
  | forall i :: 0 <= i < |xs| ==> !isClean(xs[i])


function dirtyDataToString(ds : DirtyData) : seq<string>
{
  if |ds| == 0
    then []
    else [ds[0].value] + dataToString(ds[1..])
}

// [ NOTE ]
//
// This function can be used to violate our condition that dirty data should not be printed!
//

method {:test} testDataCollection()
{
  var ds := [rawInput("a"), rawInput("b")];
  var xs := dataToString(ds);
  expect xs == ["a", "b"];
}

// [ Question ]
//
// Can you write a datatype for collecting *clean* data?

type CleanData = Data

// [ Question ]
//
// What would you do to `dirtyDataToString` to make it work for clean data only?
//
// Should you do this?
//
// How would safe wrappers help address the problem?


/*
   Scrubs will scrub a collection of dirty data clean.

   We do this in a method to highlight some benefits of subset types!
 */
method Scrubs(ds : DirtyData) returns (res : CleanData)
  // What is missing from this specification?
{
  var xs : CleanData := [];

  for i := 0 to |ds|
  {
    var c : Datum := scrub(ds[i]);

//    xs := [c] + xs;
    xs := [ds[i]] + xs;
  }
  return xs;

}

// [ Question ]
//
// Can you write a functional version of `Scrubs`, called `scrubs`?

function scrubs(ds : DirtyData) : CleanData


method {:test} testCollectingData()
{
  var ds : DirtyData := [rawInput("a"), rawInput("b")];

  var cs : CleanData := scrubs(ds);

  expect dataToString(ds) == ["a", "b"];
  expect dataToString(cs) == ["a", "b"];
}

== [ Marker ] =============================================================== */

// -- [ EOF ]

// =============================================================================
//
// Header
//
// =============================================================================

// [ NOTE ]
//
//
