||| =============================================================================
|||
||| Reasoning about Data Hygiene using Design-by-Contract
|||
||| =============================================================================
|||
||| TYPESTATE associates types with a state, and the set of operations
||| acting over the state follows a state machine.
|||
||| This file demonstrates using TYPESTATE to reasoning about data hygiene.
|||
||| We will answer some questions together, but we will give you this
||| file unmodified to try outside of the lecture.
|||
||| Remember: Lengthier, slower, notes about Dafny are on MyPlace.
|||
module Hygiene

import Decidable.Equality
import Data.List.Quantifiers

-- =============================================================================
--
-- Capturing Data Hygiene
--
-- =============================================================================

{- [ Marker ] ==================================================================

||| Represents hygiene of data coming from the outside world.
data Hygiene = CLEAN | DIRTY

--- [ NOTE ]
---
--- Hygiene is an 'enumerated' type that only has data constructors.

||| We will need to compare instances of `Hygiene` together.
|||
eq : (a : Hygiene) -> (b : Hygiene) -> Bool
eq CLEAN CLEAN = True
eq DIRTY DIRTY = True
eq _ _
  = False

-- [ Question ]
--
-- Can you rewrite `eq` to use a single equation with a single case
-- expression on the right-hand side?

Eq Hygiene where
  (==) = Hygiene.eq

-- [ NOTE ]
--
-- Idris uses interfaces for generic programming.
-- `Eq` is the interface for types that can be made equal.


||| Example proof that `CLEAN` and `DIRTY` are not equal.
noCleanDirty : CLEAN = DIRTY -> Void
noCleanDirty Refl impossible

||| Decidable euality for Hygiene states.
decEq : (a,b : Hygiene) -> Dec (a = b)

decEq CLEAN CLEAN
  = Yes Refl

decEq CLEAN DIRTY
  = No noCleanDirty

decEq DIRTY CLEAN
  = No (negEqSym noCleanDirty)

decEq DIRTY DIRTY
  = Yes Refl

||| Decidable euality for Hygiene states.
DecEq Hygiene where
  decEq = Hygiene.decEq

-- =============================================================================
--
-- Labelling Data
--
-- =============================================================================
--
-- We will now build labelled data that associates a type with a state.
--

||| Datum pairs strings with a hygiene status.
|||
||| @status indexes the type of `Datum`.
|||
||| Type indices are erased for runtime.
data Datum : (status : Hygiene)
                    -> Type
  where
    ||| Create an instance of `Datum`.
    D : (value : String) -> Datum status

-- [ NOTE ]
--
-- The data constructor for `Datum`, (`D`) does not provide a value
-- level parameter to *promote* to the type level.
--
-- This is on purpose, as we do not want the status to be dependend
-- upon at runtime!


-- [ Question ]
--
-- Can you write a version of `Datum` that has an argument in the data
-- constructor that is promoted to the type level?

-- [ Question ]
--
-- Can you write a version of `Datum` that is defined as a record?


-- =============================================================================
--
-- A State Machine for Data Hygenie
--
-- =============================================================================
--
-- We will now build labelled data that associates a type with a
-- state, and maintains that state.
--
-- Our state machine is simple:
--
-- string -[creation]-> DIRTY -[scrub]-> CLEAN -> string
--
-- That is, when raw strings enter `Data` they are labelled dirty.
-- Data remains dirty until it is cleaned. Finally we can only access
-- the underlying string it it has been cleaned,


||| A factory for labelling data dirty.
|||
||| We can use `rawInput` to place strings into a labelled datatype.
|||
rawInput : (s : String) -> Datum DIRTY
rawInput s = D s

-- [ NOTE ]
--
-- That we can just *state* that the result is dirty.

-- [ Test ]
--
-- Uncomment the next two lines!
--

testCreation : rawInput "Hello World" = D "Hello World"
testCreation = Refl

||| [ Question ]
|||
||| Can you complete the definition of 'scrub', complete with
||| specification?
|||
||| We will assume that cleaning is the following magic identity function:

magic : (s : String) -> String
magic s = s

||| Scrub takes, as input, dirty data and returns clean data.
|||
scrub : (s : Datum DIRTY) -> Datum CLEAN

-- [ Test ]
--
-- Uncomment the next two lines!
--
-- testScrub : scrub (rawInput "Hello World") = "Hello World"
-- testScrub = Refl

-- [ Question ]
--
-- Do we need the pre-condition of data?
--
-- Can you write a version of scrub that takes any Datum as input?

||| Print extracts cleaned data and displays it to the console.
|||
printLn : (d : Datum CLEAN) -> IO ()
printLn (D value) = IO.printLn value

-- =============================================================================
--
-- Collecting Hygiene
--
-- =============================================================================
--
-- `Datum` is okay, but collecting data is also important.


-- ||| A collection of Datum, Data.
-- |||
-- Data : Type
-- Data = List Datum

-- [ Question ]
--
-- This type synonym will not type check. Can you describe what is 'wrong' with this
-- definition?
--
-- Uncomment it and see what the error message says.
--

||| Dirty data is a list of dirty datum
DirtyData : Type
DirtyData = List (Datum DIRTY)

example : DirtyData
example = [rawInput "a", rawInput "b" ]

-- [ Note ]
--
-- Look at the definition of DirtyData in Dafny. What has changed?
-- What is the same?


-- [ Question ]
--
-- Can you write a definition for `CleanData`.
--

CleanData : Type

-- Uncomment the next two lines.
--
-- exampleCleanData : DirtyData
-- exampleCleanData = map scrub [rawInput "a", rawInput "b" ]

-- [ NOTE ]
--
-- This function uses a Higher Order function called `map` to `scrub`
-- the dirty data clean.

||| Extract the string values.
toString : List (Datum a) -> List String
toString xs = map (\(D v) => v) xs

-- [ NOTE ]
--
-- This function can be used to violate our condition that dirty data
-- should not be printed!
--


-- [ Test ]
--
-- Uncomment the next two lines!
--
-- testToString : toString [rawInput "Hello World"] = ["Hello World"]
-- testToString = Refl

-- [ Question ]
--
-- Can you write a definition for `toString` that uses pattern
-- matching and recursion?
--

-- [ Question ]
--
-- What would you do to `toString` to make it work for clean data
-- only?
--
-- Should you do this?
--
-- How would Idris namespaces (or modules) help address the problem?

|||  Scrubs will scrub a collection of dirty data clean.
scrubs : DirtyData -> CleanData

-- [ Question ]
--
-- Can you complete the function `scrubs`?


-- [ Test ]
--
-- Uncomment the next two lines!
--
-- testScrubs : scrubs [rawInput "Hello World"] = ["Hello World"]
-- testScrubs = Refl

-- [ EOF ]

== [ Marker ] =============================================================== -}

-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--
