// =============================================================================
//
// Security of Information
//
// =============================================================================
//
// TYPESTATE associates types with a state, and the set of operations
// acting over the state follows a state machine.
//
// This file demonstrates using TYPESTATE to reasoning about security of
// information.
//
// We will answer some questions together, but we will give you this
// file unmodified to try outside of the lecture.
//
// Remember: Lengthier, slower, notes about Dafny are on MyPlace.
//
// We also are using modules, because we are good software engineers.
// Beware when moving the marker!


/* [ Marker ] ==================================================================

// =============================================================================
//
// Encoding Security Levels
//
// =============================================================================

module SecurityLevels
{

  /**
     A classic set of labels for security.

   */
  datatype SecLevel
    = TopSecret
    | Secret
    | Public

  // [ NOTES ]
  //
  // Here we have defined a new datatype `SecLevel` to encode the levels
  //
  // `SecLevel` will have constructors `TopSecret`, `Secret`, and `Public`.
  //
  // If you have seen enumerated types, then `SecLevel` is one.
  //
  // This datatype only represents the different levels, not how they are compared.


  /**
   `equal` is a function that compares to security levels and returns True
    if they are equal, false otherwise.
   */
  function equal(a : SecLevel, b : SecLevel) : bool
  {
    (a == TopSecret && b == TopSecret)
      || (a == Secret && b == Secret)
      || (a == Public && b == Public)
  }

  // [ NOTES ]
  //
  // Equal is great, but we need more if we are going to attest about levels.
  //
  // We can do so using a *Comparison* function.
  //
  // First we capture the result of a comparison.


  /*
     Ordering captures the result of comparing data.

     Given two values, they can be:

     + `LT` less than
     + `EQ` equal to
     + `GT` greater than

   */
  datatype Ordering = LT | EQ | GT


 /*
    Compares the security level `a` against `b`, returning a result from the perspective of `a`.

    Compare realises our Security Heirachy.

    Specifically, the following heirachy.

    ```
    Public < Secret < TopSecret
    ```
 */
 function compare(a : SecLevel, b : SecLevel) : Ordering
 {
  match a
  {
  case TopSecret =>   // Top Secret it the highest
    if b == TopSecret
      then EQ
    else GT
  case Secret =>     // Next is Secret
    match b
    {
    case TopSecret => LT
    case Secret => EQ
    case Public => GT
    }
    case Public =>  /// Finally is Public
      if b == TopSecret || b == Secret
        then LT
      else EQ
  }
 }

 // [ Question ]
 //
 // The definition of `compare` is very complicated for Dafny.
 //
 // Can you rewrite the comparison in a better way?
}


// =============================================================================
//
// Tagged Information
//
// =============================================================================
module TaggedInformation
{
  import opened SecurityLevels // <- import all of SecurityLevels

  /**
     Info is a datatype we can use to associate a value with a security level.
   */
  datatype Info<T> = I(value : T,  level : SecLevel)

  // [ Question ]
  //
  // Can you think why we want to have the security level not be a
  // ghost specification?

  function topsecret<T>(x : T) : Info<T>
  {
    I(x,TopSecret)
  }

  function secret<T>(x : T) : Info<T>
  {
    I(x,Secret)
  }

  function public<T>(x : T) : Info<T>
  {
    I(x,Public)
  }

  predicate isSecret<T>(i : Info<T>)
  {
    compare(i.level, Secret) == EQ
  }

  // [ Question ]
  //
  // Is `isSecret` correct?
  //
  // If you think not, why? How coudl we fix it?

  method {:test} testIsSecret()
  {

    expect isSecret(topsecret(1));
    expect isSecret(secret(1));
    expect !isSecret(public(1));

  }

  // [ Question ]
  //
  // Fix `isSecret`?


  // ===========================================================================
  //
  // Working with Tagged Information.
  //
  // ===========================================================================
  //
  // Here we will show some examples of working with `Info`.
  //

  /*
     Print is a method that will only print public information.
   */
  method Print(i : Info<string>)
    requires i.level == Public
  {
    print i.value, "\n";
  }

  // [ Question ]
  //
  // Can you rewrite the post condition using a better predicate?
  //


  function declassify<T>(i : Info<T>) : Info<T>
    requires compare(i.level, Public) == GT
    ensures  declassify(i).level == Public
  {
    i.(level := Public)
  }


  /*
     An example of using `Print`
   */
  method {:test} testTaggedInfo()
  {
    var ts_val := topsecret("Exam Answers");
    var p_val := public("Exam Questions");
    Print(p_val);
    // Print(ts_val);
    // ^ This fails as the variable `ts_val`, is not public
  }


}

// ===========================================================================
//
// Working with Tagged Information 2
//
// ===========================================================================
//
// Here we will show some examples of working with `Info`.
//
module TaggedCollections
{
  import opened SecurityLevels
  import opened TaggedInformation

  type List<T> = seq<Info<T>>

  // [ Question ]
  //
  // What is wrong with this definition?


  method {:test} testList()
  {
    var xs : List<string> :=
      [ topsecret("Exam Answers")
      , public("Exam Questions")
      ];

    for i := 0 to |xs|
    {
//      Print(xs[i]);
    }
  }

  // [ Question ]
  //
  // Can you make the test run?


  /**
     Here we have used a type synonym, `PublicStrings`, to declare a sequence
     of public strings as a subset type of a sequence of Info.

     This type synonym is special as we use a 'constraint' to reason about
     the type's contents
   */
  type PublicThings<T> = xs : seq<Info<T>>                              // <- This is the original type
                       | forall i :: 0 <= i < |xs| ==> xs[i].level == Public // <- This is the constraint

  // [ NOTE ]
  //
  // The constraint states that all elements in the sequence must be
  // public.

  method Use2()
  {
    var ts_val := topsecret("Exam Answers");
    var p_val := public("Exam Questions");

    var pxs : PublicThings<string> := [p_val, p_val];

    // var sxs : PublicStrings := [ts_val, p_val];
    // ^ fails because of `ts_val`.

    var i := 0;
    for i := 0 to |pxs| - 1
    {
      Print(pxs[i]);
    }
  }

  type SecretThings<T> = xs : seq<Info<T>> | forall i :: 0 <= i < |xs| ==> isSecret(xs[i])

  // [ Question ]
  //
  // Can you write a sequence of Tagged Information for all secret information.

  method Use3()
  {
    var ts_val := topsecret("Exam Answers");
    var s_val := secret("Exam Answers");
    var p_val := public("Exam Questions");

    var sxs : SecretThings<string> := [ts_val, s_val];
    // ^ This should not fail.
    var pxs : PublicThings<string> := [ts_val, p_val];
    // ^ this fails because of `ts_val`.

    var i := 0;
    for i := 0 to |pxs|
    {
      Print(pxs[i]);
    }
  }
}

== [ Marker ] =============================================================== */

// -- [ EOF ]

// =============================================================================
//
// Header
//
// =============================================================================

// [ NOTE ]
//
//

// [ Question ]
//
//
