-- =============================================================================
|||
||| Security of Information
|||
-- =============================================================================
|||
||| TYPESTATE associates types with a state, and the set of operations
||| acting over the state follows a state machine.
|||
||| This file demonstrates using TYPESTATE to reasoning about security of
||| information.
|||
||| We will answer some questions together, but we will give you this
||| file unmodified to try outside of the lecture.
|||
||| Remember: Lengthier, slower, notes about Dafny are on MyPlace.
|||
||| We will be using namespaces, because we are good proof engineers
module Information

import Decidable.Equality
import Data.List.Quantifiers

-- ^ We will need these later on

{- [ Marker ] ==================================================================

--  ============================================================================
|||
||| Encoding Security Levels
|||
-- ============================================================================
|||
namespace SecurityLevels

  ||| A classic set of labels for security.
  public export
  data SecLevel
    = TopSecret
    | Secret
    | Public


  --  [ NOTES ]
  --
  -- Here we have defined a new datatype `SecLevel` to encode the
  -- levels
  --
  -- `SecLevel` will have constructors `TopSecret`, `Secret`, and
  -- `Public`.
  --
  -- If you have seen enumerated types, then `SecLevel` is one.
  --
  -- This datatype only represents the different levels, not how they
  -- are compared.
  --

  ||| `equal` is a function that compares to security levels and
  ||| returns True if they are equal, false otherwise.
  |||
  |||
  equal : (a,b : SecLevel) -> Bool

  -- [ Note ]
  --
  -- Top Secret is equal to itself
  equal TopSecret TopSecret
    = True

  -- [ Note ]
  --
  -- Anything else is false
  equal TopSecret _ -- <- `_` underscore represents the wildcard
    = False

  -- [ Note ]
  --
  -- Same as with top secret
  equal Secret Secret
    = False
  equal Secret _
    = False

  -- [ Note ]
  --
  -- Same as with top secret
  equal Public Public
    = True
  equal Public _
    = False

  ||| `Eq` is a interface in Idris we can provide an instance for
  ||| SecLevel so that we can use SecLevel where the `Eq` interface is
  ||| expected.
  export
  Eq SecLevel where
    (==) = equal

  -- [ NOTES ]
  --
  -- Whilst `Eq` is fine, let us do something stronger!

  decEq : (a,b : SecLevel) -> Dec (Equal a b)

  -- Some things that are just not true!

  Uninhabited (Equal TopSecret Public) where
    uninhabited Refl impossible

  Uninhabited (Equal TopSecret Secret) where
    uninhabited Refl impossible

  Uninhabited (Equal Secret Public) where
    uninhabited Refl impossible

  decEq TopSecret TopSecret
    = Yes Refl

  decEq TopSecret Secret
    = No absurd

  decEq TopSecret Public
    = No absurd

  decEq Secret TopSecret
    = No (negEqSym absurd)

  decEq Secret Secret
    = Yes Refl

  decEq Secret Public
    = No absurd
  decEq Public TopSecret
    = No (negEqSym absurd)

  decEq Public Secret
    = No (negEqSym absurd)

  decEq Public Public
    = Yes Refl

  export
  DecEq SecLevel where
    decEq = SecurityLevels.decEq

  -- [ NOTES ]
  --
  -- Equal is great, but we need more if we are going to attest about levels.
  --
  -- We can do so using a *Comparison* function.
  --
  -- First we capture the result of a comparison.
  --

  ||| Ordering is predefined in Idris2 so we are hiding it.
  namespace Hidden

    ||| Ordering captures the result of comparing data.
    |||
    ||| Given two values, they can be:
    |||
    ||| + `LT` less than
    ||| + `EQ` equal to
    ||| + `GT` greater than
    data Ordering = LT | EQ | GT


  ||| Compares the security level `a` against `b`, returning a result
  ||| from the perspective of `a`.
  |||
  ||| Compare realises our Security Heirachy.
  |||
  ||| Specifically, the following heirachy.
  |||
  ||| ```
  ||| Public < Secret < TopSecret
  ||| ```
  compare : (a,b : SecLevel) -> Ordering
  compare TopSecret TopSecret = EQ
  compare TopSecret _ = LT

  compare Secret TopSecret = LT
  compare Secret Secret = EQ
  compare Secret Public = GT

  compare Public Public = EQ
  compare Public _ = LT

  ||| `Ord` is a interface in Idris we can provide an instance for
  ||| SecLevel so that we can use SecLevel where the `Ord` interface
  ||| is expected.
  export
  Ord SecLevel where
    compare = SecurityLevels.compare


  -- [ NOTE ]
  --
  -- Idris does not have a type to capture decidable ordering.
  --
  -- So we will have to do this ourselves.

  ||| Is security level a greater than security level b?
  public export
  data GreaterThan : (a,b : SecLevel) -> Type where
    GThanTS : GreaterThan TopSecret Secret
    GThanTP : GreaterThan TopSecret Public
    GThanSP : GreaterThan Secret    Public

  export
  greaterThan : (a,b : SecLevel) -> Dec (GreaterThan a b)

  ||| Is security level a greater than security level b?
  public export
  data LessThan : (a,b : SecLevel) -> Type where
    LThanPS : LessThan Public Secret
    LThanPT : LessThan Public TopSecret
    LThanTS : LessThan Secret TopSecret

  export
  lessThan : (a,b : SecLevel) -> Dec (GreaterThan a b)


  -- [ Question ]
  --
  -- Can you provide the implementations to `greaterThan` and `lessThan`?

  -- [ Question ]
  --
  -- Can you provide the definitions and implementations that capture:
  --
  -- + Greater to or equal to
  -- + Less than or equal to

-- =============================================================================
|||
||| Tagged Information
|||
-- =============================================================================
|||
namespace TaggedInformation


  ||| Info is a datatype we can use to associate a value with a security level.
  export
  data Info : (type : Type) -> SecLevel -> Type where
    I : (value : type)
              -> Info type level

  export
  topsecret : (value : type) -> Info type TopSecret
  topsecret v = I v

  export
  secret : (value : type) -> Info type Secret
  secret v = I v

  export
  info : (value : type) -> Info type Public
  info = I

  export
  declassify : (i : Info type level)
                 -> Info type Public
  declassify (I value) = I value


  -- [ Question ]
  --
  -- Is the type of `declassify` correct?
  --
  -- Do we need to be more precise or less precise?

  -- [ Question ]
  --
  -- Can you rewrite declassify in two different ways:
  --
  -- 1. Only information secret and above can be declassified with an explicit runtime proof.
  -- 2. Only information secret and above can be declassified using an implicit proof.
  --

  export
  printLnInfo : Show a
             => (i : Info a Public)
                  -> IO ()
  printLnInfo (I value) = printLn value


  use1 : IO ()
  use1 = do let ts_val = topsecret "Exam Answers"
            let p_val  = info "Exam Questions"
            printLnInfo p_val

            -- printLnInfo ts_val
            -- ^ fails the variable `ts_val`, is not public


--  ============================================================================
|||
||| Working with Tagged Information 2
|||
--  ============================================================================
|||
namespace TaggedCollections

  ||| Here we have used a type synonym, `PublicStrings`, to declare a
  ||| sequence of public things.
  |||
  public export
  PublicThings : Type -> Type
  PublicThings type
    = List (Info type Public)

  -- [ Question ]
  --
  -- What do you notice about this declaration compared to the one in Dafny?

  use2 : IO ()
  use2 = do let ts_val = topsecret "Exam Answers"
            let p_val  = info "Exam Questions"

            let pxs : PublicThings String = [p_val, p_val]

            -- let sxs : PublicStrings String = [ts_val, p_val]
            -- ^ fails because of `ts_val`

            traverse_ printLnInfo pxs

            -- Uncomment this line
            -- traverse_ printLnInfo (the (List (Info String _)) [ts_val])

  ||| A TypeSynonm to capture a list of information that is secret and above.
  |||
  ||| We have used a dependent pair to provide evidence, adn `All` to
  ||| collect the different levels of the items in the list.
  public export
  SecretStrings : Type
  SecretStrings
    = (lvls : List SecLevel ** All (\lvl => (Info String lvl, GreaterThan Public lvl)) lvls)

  -- [ Question ]
  --
  -- Can you describe why the dependent pair was used here?
  --

  use3 : IO ()
  use3 = do let ts_val = topsecret "Exam Answers"
            let s_val  = secret "Exam Answers"
            let p_val  = info "Exam Questions"

            let pxs : PublicThings String = [p_val, p_val]
            let sxs : SecretStrings  = ([Public, Public] ** [?x, ?y])

            -- [ Question ]
            --
            -- Can you fill the holes `x` and `y`

            traverse_ printLnInfo pxs

  ||| `SecretStringsAll` demonstrates how we can provide a concrete
  ||| datatype to represent `SecretStrings`.
  |||
  ||| This demonstrating further the similarities between dependent
  ||| pairs and dependent types
  public export
  data SecretStringsAlt : Type where
    SS : All (\lvl => (Info String lvl, Equal Secret Public)) lvls
      -> SecretStringsAlt

  -- [ Question ]
  --
  -- Describe two things wrong with the definition of `SecretStringsAlt`?
  --
  -- Can you fix it?
  --

  ||| Infos is a datatype that collects infomation that is above a
  ||| security level.
  |||
  ||| `Infos` does the same job as both `ScretStrings` and
  ||| `SecretStringsAlt` but in a different way.

  data Infos : (type : Type) -> (lvl : SecLevel) -> Type where
    Empty  : Infos type lvl
    Extend : Info type lvla
          -> GreaterThan lvla lvl
          -> Infos type lvl
          -> Infos type lvl


  -- [ Question ]
  --
  -- Can you think about the advantages and disadvantages of each approach?
  --

  -- [ Question ]
  --
  -- Can you rewrite `Infos` so that it is parametric (i.e. generic)
  -- over the relation between the given security level and that of
  -- the level of the item itself?

== [ Marker ] =============================================================== -}

-- [ EOF ]
