// =============================================================================
//
// A Raw Sockets API
//
// =============================================================================
//
// Dafny does not support socket based communication nor linear types.
//
// `SocketsRaw` is an abridge version of a 'Raw' sockets API, adapted
// from Idris' `network` package, that is unrestricted in how it can
// be used.
//
// We have purposefully abstracted over the implementation details as
// our focus here is more how we can use the API not how the
// underlying system calls work.
//

// Imports!
include "Common.dfy"

import opened Common.Result
import opened Common.Maybe

// =============================================================================
//
// Header
//
// =============================================================================
//
// First, we define some datatype definitions to represent important
// networking datatypes.
//
// We have left their definition abstract on purpose.


datatype Socket = S()
datatype SocketError = E()

// =============================================================================
//
// The API Itself
//
// =============================================================================
//
// We now define our API that abstracts over a lower-level API taken
// from a systems-level library.

/** Creates a 'socket'.
 */
method NewSocket() returns (res : Result<SocketError,Socket>)

/**
  Tears it all down
 */
method Close(sock : Socket)

// =============================================================================
//
// Creating Server Sockets
//
// =============================================================================


/**
 Binds a 'socket' to the given address and `port`.

 Returns `0` on success, or an error code.
*/
method Bind(sock : Socket, addr : Maybe<string>, port : nat) returns (res : int)

/**
 Start listening for incoming connections.

 Returns `0` on success, or an error code.
*/
method Listen(sock : Socket) returns (res : int)

/**
 Accept a connection on the provided socket.

 Returns on failure a `SocketError`
 Returns on success a pairing of:

 + `Socket`        :: The socket representing the connection.
 + `SocketAddress` :: The address of the incoming connection.
*/
method Accept(sock : Socket) returns (res : Result<SocketError,(Socket,string)>)


// =============================================================================
//
// Creating Connections
//
// =============================================================================


/**
 Connects to a given address and port.

 Returns 0 on success, and an error number on error.
*/
method Connect(sock : Socket, addr : string, port : nat) returns (res : int)

/**
 Send the string message.

 Returns on failure a `SocketError`.
 Returns on success the `ResultCode`.
*/
method Send(sock : Socket, addr : string) returns (res : Result<SocketError,int>)

/**
 Receive data on the specified socket.

 Returns on failure a `SocketError`
 Returns on success a pairing of:

 + `String`     :: The payload.
 + `ResultCode` :: The result of the underlying function.
*/
method Receive(sock : Socket, len : nat) returns (res : Result<SocketError,(string, int)>)


// =============================================================================
//
// A Bad Program
//
// =============================================================================
//
method Prog()
{
  var socke := NewSocket();
  match socke
  {
  case Error(e) => print "Error\n";
  case Success(sock) =>
    var res := Connect( sock, "localhost", 8423);
    if res == 0
    { // This is the good branch!
      print "Failure\n";
      return;
    }

    var resS := Send(sock, "Ping");
    match resS
    {
    case Error(err) => print "Error\n";
    case Success(m) => return;
    }
  }
}  // None of the sockets have been closed.

/* [ Marker ] ==================================================================
== [ Marker ] =============================================================== */

// -- [ EOF ]

// =============================================================================
//
// Header
//
// =============================================================================

// [ NOTE ]
//
//

// [ Question ]
//
//
