--- =============================================================================
---
--- Dependent Linear TYPESTATE: Mixing Linear Types and TYPESTATE, and
--- handling errors
---
--- =============================================================================
|||
||| We build upon the continuation style linear and TYPESTATE sockets
||| API, so that errors are handled.
|||
||| For brevity we only look at the operations from a Client's perspective.
|||
||| The exercises will ask you to complete the server variant.
|||
||| The `network` library, packaged as part Idris' standard libraries,
||| provides a working Linear Raw network implementation that you can play
||| with, not to mention a linear one.
|||
|||
module Sockets

-- =============================================================================
--
-- Useful datatypes
--
-- =============================================================================
--
-- Rather than define a complete set of states for both server and client sockets,
-- we are concentrating on the client operations.
--
-- As with all the code, We have left some definitions abstract on purpose.

data SState = READY | OPEN | CLOSED

data Socket : SState -> Type where

export
data Error : Type where

-- =============================================================================
--
-- State Changing
--
-- =============================================================================
--
-- We need to map actions to states, and determine the `NextSteps` in
-- the finite state machine.
--

||| `ClientAction` maps the expected state of the socket, with an
||| action.
data ClientAction : SState -> Type where
  Connect : ClientAction READY
  Send    : ClientAction OPEN
  Receive : ClientAction OPEN
  Close   : ClientAction st


||| We need to track if the output of an error represents a positive or negative result.
|||
||| When using `Maybe`, `Nothing` can sometimes represent a good thing!
data Polarity = POS | NEG

fromMaybe : Maybe a -> Polarity
fromMaybe Nothing = NEG
fromMaybe (Just x) = POS

fromMaybe' : Maybe a -> Polarity
fromMaybe' Nothing = POS
fromMaybe' (Just x) = NEG

fromEither : Either a b -> Polarity
fromEither (Left x) = NEG
fromEither (Right x) = POS

||| `NextStep` calculates the next type of the socket, based on a given polarity.
|||
||| Positive outputs perform the state change to the next positive step.
||| Negative outputs (i.e. errors) changes the socket to closed.
|||
||| We calculate a type rather than a `SState` instance, because of
||| where/how we perform the calculation.
NextStep : (act : ClientAction st)
        -> (pol : Polarity)
               -> Type
NextStep Connect POS
  = Socket OPEN

NextStep Send POS
  = Socket OPEN

NextStep Receive POS
  = Socket OPEN

NextStep Close POS
  = Socket CLOSED

NextStep _ NEG
  = Socket CLOSED



-- =============================================================================
--
-- The API itself
--
-- =============================================================================
--
-- We now define our API that abstracts over a lower-level API taken
-- from a systems-level library.
--

||| Creates a 'socket'.
|||

socket : (cont : (1 sock : Socket READY) -> IO ())
              -> IO ()

||| Tears it all down
export
close : (1 sock : Socket st) -> IO ()



-- =============================================================================
--
-- Creating Connections
--
-- =============================================================================

||| Connects to a given address and port.
|||
||| Returns Nothing on success, or Just an error.
|||
export
connect : (1 sock : Socket READY)
       -> (  addr : String)
       -> (  port : Nat )
       -> (  cont : (1 res : Res (Maybe Error)
                                 (\res => NextStep Connect (fromMaybe' res)))
                          -> IO ())

                 -> IO ()

-- [ NOTE ]
--
-- `Res` is a dependent pair in which the first element is linear.
--
-- We give its definition here for convenience
--
namespace Hidden
  data MyRes : (a : Type) -> (prf : a -> Type) -> Type where
    MkRes : (val : a) -> (1 r : t val) -> MyRes a t

||| Send the string message.
|||
||| Returns on failure a `SocketError`.
||| Returns on success the `ResultCode`.
|||
export
send : (1 sock : Socket OPEN)
    -> (  msg  : String)
    -> (  cont : (1 res : Res (Maybe Error)
                              (\res => NextStep Send (fromMaybe' res)))
                       -> IO ())
              -> IO ()

||| Receive data on the specified socket.
|||
||| Returns on failure a `SocketError`
||| Returns on success a pairing of:
|||
||| + `String`     :: The payload.
||| + `ResultCode` :: The result of the underlying function.
|||
export
recv : (1 sock : Socket OPEN)
    -> (  len  : Nat)
    -> (  cont : (1 res : Res (Either Error (String, Int))
                              (\res => NextStep Receive (fromEither res)))
                           -> IO ())
              -> IO ()


{- The exercise sheet will ask you to rewrite this to handle errors!
-- =============================================================================
--
-- Creating Server Sockets
--
-- =============================================================================

||| Binds a 'socket' to the given address and `port`.
|||
||| Returns `0` on success, or an error code.
|||
export
bind : (1 sock  : Socket READY)
    -> (  addr  : Maybe String)
    -> (  port  : Nat)
    -> (  cnt   : (1 sock : Socket BOUND) -> IO ())
               -> IO ()

||| Start listening for incoming connections.
|||
||| Returns `0` on success, or an error code.
|||
export
listen : (1 sock : Socket BOUND)
      -> (  cnt  : (1 sock : Socket LISTEN) -> IO ())
                -> IO ()

||| Accept a connection on the provided socket.
|||
||| Returns on failure a `SocketError`
||| Returns on success a pairing of:
|||
||| + `Socket`        :: The socket representing the connection.
||| + `SocketAddress` :: The address of the incoming connection.
|||
export
accept : (1 sock : Socket LISTEN)
      -> (  cnt  : (1 res : LPair (Socket OPEN) String) -> IO ())
      -> IO ()
-}


-- =============================================================================
--
-- A Bad Program
--
-- =============================================================================
--

-- [ Question ]
--
-- Can you rewrite the program so that it works correctly.

prog : IO ()
prog
  = do socke <- socket
       case socke of
         Left err => putStrLn "Error"
         Right sock
           => do res <- connect sock "localhost" 8423
                 if res == 0
                  -- This is actually the good branch!
                  then putStrLn "Failure"
                  -- This is the failing branch
                  else do res <- send sock "Ping"
                          case res of
                            Left err => putStrLn "Error"
                            Right sock => pure ()
                  -- None of the sockets have been closed.

-- [ EOF ]


-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--

{- [ Marker ] ==================================================================

== [ Marker ] =============================================================== -}
