--- =============================================================================
---
--- Mixing Linear Types and TYPESTATE, a first attempt.
---
--- =============================================================================
|||
||| `SocketsLinear` builds ontop of the TYPESTATE API by using QTT to
||| ensure that instances of socket can only be ued once.
|||
||| Once again, we have purposefully abstracted over the
||| implementation details as our focus here is more how we can use
||| the API not how the underlying system calls work.
|||
||| Sequencing linear data when programming functionally is not
||| straightforward.  As you have seen with the 'Raw' and `TYPESTATE`
||| APIs we have leveraged the `IO` computation context, which enables
||| us to sequence interactions with the outside world using 'Do'
||| notation.
|||
||| Unfortunatly, we cannot mix `IO` and linearity directly as the
||| core operations within the `IO` computation context introduces
||| unrestricted values. Take, for example, the following function
||| type signature represents sequencing:
|||
||| ```idris
||| seq : (this : IO a) -> (thenThis : a -> IO b) -> IO b
||| ```
|||
||| Here the implementation of `seq` will first run the `this`
||| computation and then pass the result into the next computation (`thenThis`).
|||
||| The linear version is:
|||
||| ```idris
||| seq : (1 this : IO a) -> (thenThis : (1 res : a) -> IO b) -> IO b
||| ```
|||
||| Using linearity/QTT in idris is not straightforward the `linear`
||| package, shipped as part of Idris standard libraries, provides
||| more detailed information about how we can work with quantities
||| within Idris.
|||
||| + Note ::
|||   We will not expect nor ask you to use, or understand,
|||   `IO` nor `LIO`.
|||
||| To make working with linear values more accessible, we have
||| decided to use continuation passing to sequence the
||| computations. There is a video on MyPlace from Week 4 that
||| demonstrates this.
|||
||| Moreover, we have designed the API to:
|||
||| 1. not handle errors;
||| 2. nor perform error based state changes;
|||
||| We will address this in live coding in the second half of this
||| lecture.
|||
||| The `network` library, packaged as part Idris' standard libraries,
||| provides a working Raw network implementation that you can play
||| with, not to mention a linear one.
|||
|||
module SocketsLinear

-- =============================================================================
--
-- Useful datatypes
--
-- =============================================================================
--
-- First, we define some datatype definitions to represent important
-- networking datatypes.
--
-- We have left their definition abstract on purpose.

data State = READY | OPEN | CLOSED | BOUND | LISTEN

data Socket : State -> Type where

export
data Error : Type where

-- =============================================================================
--
-- The API itself
--
-- =============================================================================
--
-- We now define our API that abstracts over a lower-level API taken
-- from a systems-level library.
--

||| Creates a 'socket'.
|||
export
socket : (cont : (1 sock : Socket READY) -> IO ())
              -> IO ()

||| Tears it all down
export
close : (1 sock : Socket st) -> IO ()


-- =============================================================================
--
-- Creating Server Sockets
--
-- =============================================================================

||| Binds a 'socket' to the given address and `port`.
|||
||| Returns `0` on success, or an error code.
|||
export
bind : (1 sock  : Socket READY)
    -> (  addr  : Maybe String)
    -> (  port  : Nat)
    -> (  cnt   : (1 sock : Socket BOUND) -> IO ())
               -> IO ()

||| Start listening for incoming connections.
|||
||| Returns `0` on success, or an error code.
|||
export
listen : (1 sock : Socket BOUND)
      -> (  cnt  : (1 sock : Socket LISTEN) -> IO ())
                -> IO ()

||| Accept a connection on the provided socket.
|||
||| Returns on failure a `SocketError`
||| Returns on success a pairing of:
|||
||| + `Socket`        :: The socket representing the connection.
||| + `SocketAddress` :: The address of the incoming connection.
|||
export
accept : (1 sock : Socket LISTEN)
      -> (  cnt  : (1 res : LPair (Socket OPEN) String) -> IO ())
      -> IO ()

-- =============================================================================
--
-- Creating Connections
--
-- =============================================================================

||| Connects to a given address and port.
|||
||| Returns 0 on success, and an error number on error.
|||
export
connect : (1 sock : Socket READY)
       -> (  addr : String)
       -> (  port : Nat )
       -> (  cnt  : (1 sock : Socket OPEN) -> IO ())
                 -> IO ()

||| Send the string message.
|||
||| Returns on failure a `SocketError`.
||| Returns on success the `ResultCode`.
|||
export
send : (1 sock : Socket OPEN)
    -> (  msg  : String)
    -> (  cnt  : (1 res : Socket OPEN) -> IO ())
              -> IO ()

||| Receive data on the specified socket.
|||
||| Returns on failure a `SocketError`
||| Returns on success a pairing of:
|||
||| + `String`     :: The payload.
||| + `ResultCode` :: The result of the underlying function.
|||
export
recv : (1 sock : Socket OPEN)
    -> (  len  : Nat)
    -> (  cnt  : (1 res : Socket OPEN) -> IO ())
              -> IO ()


-- =============================================================================
--
-- A Bad Program
--
-- =============================================================================
--

-- [ Question ]
--
-- Can you rewrite the program so that it works correctly.

prog : IO ()
prog
  = do socke <- socket
       case socke of
         Left err => putStrLn "Error"
         Right sock
           => do res <- connect sock "localhost" 8423
                 if res == 0
                  -- This is actually the good branch!
                  then putStrLn "Failure"
                  -- This is the failing branch
                  else do res <- send sock "Ping"
                          case res of
                            Left err => putStrLn "Error"
                            Right sock => pure ()
                  -- None of the sockets have been closed.

-- [ EOF ]


-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--

{- [ Marker ] ==================================================================

== [ Marker ] =============================================================== -}
