--- =============================================================================
---
--- TYPESTATE and Sockets
---
--- =============================================================================
|||
||| `SocketsTypeState` demonstrates how we can write a better sockets
||| API by annotating the `Socket` datatype with the potential states
||| that a socket can be in. We have also provided some better error
||| handling.
|||
||| With this 'typed' API we can provide better abstractions over
||| system bindings to low level networking primitives.
|||
|||
||| We have purposefully abstracted over the implementation details as
||| our focus here is more how we can use the API not how the
||| underlying system calls work.
|||
||| The `network` library, packaged as part Idris' standard libraries,
||| provides a working Raw network implementation that you can play
||| with.
|||
|||
module SocketsTypeState

-- =============================================================================
--
-- Useful datatypes
--
-- =============================================================================
--
-- First, we define some datatype definitions to represent important
-- networking datatypes.
--
-- We have left their definition abstract on purpose.

data State = READY | OPEN | CLOSED | BOUND | LISTEN

data Socket : State -> Type where

data Error : Type where

-- =============================================================================
--
-- The API itself
--
-- =============================================================================
--
-- We now define our API that abstracts over a lower-level API taken
-- from a systems-level library.
--

||| Creates a 'socket'.
|||
export
socket : IO (Either Error (Socket READY))

||| Tears it all down
export
close : Socket st -> IO ()


-- =============================================================================
--
-- Creating Server Sockets
--
-- =============================================================================

||| Binds a 'socket' to the given address and `port`.
|||
||| Returns `0` on success, or an error code.
|||
export
bind : Socket READY -> Maybe String -> Nat -> IO (Either Error (Socket BOUND))

||| Start listening for incoming connections.
|||
||| Returns `0` on success, or an error code.
|||
export
listen : Socket BOUND -> IO (Either Error (Socket LISTEN))

||| Accept a connection on the provided socket.
|||
||| Returns on failure a `SocketError`
||| Returns on success a pairing of:
|||
||| + `Socket`        :: The socket representing the connection.
||| + `SocketAddress` :: The address of the incoming connection.
|||
export
accept : Socket LISTEN
      -> IO (Either Error (Socket OPEN,String))

-- =============================================================================
--
-- Creating Connections
--
-- =============================================================================

||| Connects to a given address and port.
|||
||| Returns 0 on success, and an error number on error.
|||
export
connect : Socket READY -> String -> Nat -> IO (Either Error (Socket OPEN))

||| Send the string message.
|||
||| Returns on failure a `SocketError`.
||| Returns on success the `ResultCode`.
|||
export
send : Socket OPEN -> String -> IO (Either Error Int)

||| Receive data on the specified socket.
|||
||| Returns on failure a `SocketError`
||| Returns on success a pairing of:
|||
||| + `String`     :: The payload.
||| + `ResultCode` :: The result of the underlying function.
|||
export
recv : Socket OPEN -> Nat -> IO (Either Error (String, Int))


-- =============================================================================
--
-- A Bad Program
--
-- =============================================================================
--

-- [ Question ]
--
-- Can you rewrite the program so that it works correctly.

prog : IO ()
prog
  = do socke <- socket
       case socke of
         Left err => putStrLn "Error"
         Right sock
           => do res <- connect sock "localhost" 8423
                 if res == 0
                  -- This is actually the good branch!
                  then putStrLn "Failure"
                  -- This is the failing branch
                  else do res <- send sock "Ping"
                          case res of
                            Left err => putStrLn "Error"
                            Right sock => pure ()
                  -- None of the sockets have been closed.

-- [ EOF ]


-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--

{- [ Marker ] ==================================================================

== [ Marker ] =============================================================== -}
