--- =============================================================================
---
--- Naive Implementation of Binary Session Types
---
--- =============================================================================
|||
||| This file presents two naive implementations of binary session types
||| using an Embedded Domain Specific Language approach.
|||
||| For simplicity, we will *not* realise recursive actions, which
||| require more thought *and* engineering.
|||
||| The two approaches are:
|||
||| 1. Using Raw Types that restrictes reasoning;
||| 2. Using a Universe Pattern to enable more reasoning;
|||
module SessionsBinarySimple

import Decidable.Equality

-- =============================================================================
--
||| Simple Approach to Capturing Local Actions
--
-- =============================================================================
namespace Simple

  ||| `Action` captures the communication actions that an endpoint will
  ||| be required to perform.
  |||
  ||| `Action` has been encoded as a continuation to ensure a normal
  ||| form and that a series of actions are guaranteed to terminate
  ||| i.e. Stop!
  |||
  public export
  data Action : Type where

    ||| Send a message of type `payloadType`, and then do the 'next'
    ||| action.
    |||
    Send : (payloadType : Type)
        -> (next        : Action)
                       -> Action

    ||| Receive a message of type `payloadType`, and then do the 'next'
    ||| action.
    |||
    Recv : (payloadType : Type)
        -> (next        : Action)
                       -> Action

    ||| Terminate the interaction
    Stop : Action


  -- [ Question ]
  --
  -- Can you think of a way to reduce the number of constructors to just two.

  ||| Compute the 'dual' of an action.
  Dual : Action -> Action
  Dual (Send x y) = Recv x (Dual y)
  Dual (Recv x y) = Send x (Dual y)
  Dual Stop = Stop

  -- [ NOTE ]
  --
  -- Compare `Dual` with the definition in the lecture slides.

  EchoClient : Action
  EchoClient
    = Send String
    $ Recv String
      Stop

  EchoServer : Action
  EchoServer
    = Dual (EchoClient)

  -- [ QUESTION ]
  --
  -- After loading this file, evaluate `EchoServer` in the REPL.
  --
  -- What is the output, and how does it compare to `EchoClient`.

-- =============================================================================
--
||| Using Universes to Better Reason about Local Actions
--
-- =============================================================================
---
|||
||| Our `Simple.Action` implementation is a good start, however, Idris
||| does not allow pattern matching nor equality of Types. As such we
||| cannot reason about `Simple.Action` using `Dec`.
|||
||| To address the limitations of `Simple.Action` we need to employ
||| the universe pattern: Representing raw Types using a datatype, and
||| mapping them to concrete types later on.
|||
namespace Better

  ||| `Action` captures the communication actions that an endpoint will
  ||| be required to perform.
  |||
  ||| We have indexed `Action` by a type that we will use to represent
  ||| messages.
  |||
  ||| `Action` has been encoded as a continuation to ensure a normal
  ||| form and that a series of actions are guaranteed to terminate
  ||| i.e. Stop!
  |||
  data Action : (payloadTypes : Type) -> Type where

    ||| Send a message of type `payloadType`, and then do the 'next'
    ||| action.
    |||
    Send : (payloadType : a)
        -> (next        : Action a)
                       -> Action a

    ||| Receive a message of type `payloadType`, and then do the 'next'
    ||| action.
    |||
    Recv : (payloadType : a)
        -> (next        : Action a)
                       -> Action a

    ||| Terminate the interaction
    Stop : Action a


  -- [ Question ]
  --
  -- Can you think of a way to reduce the number of constructors to just two.

  ||| Compute the 'dual' of an action
  Dual : Action a -> Action a
  Dual (Send x y) = Recv x (Dual y)
  Dual (Recv x y) = Send x (Dual y)
  Dual Stop = Stop

  -- [ NOTE ]
  --
  -- The definition of `Dual` is the same as before.


  ||| Represents `Duality` as a datatype.
  |||
  ||| `AreDual` is our theorem that describes more formally what `Duality` is!
  |||
  ||| By representing message types as an external datatype (the
  ||| Universe Pattern) we can then perform decidable equality on
  ||| message types.
  |||
  data AreDual : (this, that : Action a) -> Type where

    ||| The Dual of `Stop` is `Stop`.
    DStop : AreDual Stop Stop

    ||| The dual of `Send` is a `Recv` on the same message type,
    ||| together with the dual of the continuation.
    DSend : (prfEQ : Equal x y)
         -> (cont : AreDual this that)
                 -> AreDual (Send x this) (Recv y that)

    ||| The dual of `Recv` is a `Send` on the same message type,
    ||| together with the dual of the continuation.
    DRecv : (prfEQ : Equal x y)
         -> (cont  : AreDual this that)
                  -> AreDual (Recv x this) (Send y that)

  -- [ Question ]
  --
  -- Can you think of a way to reduce the number of constructors to just two?


  -- Some `Uninhabited` instances to declare things that are not duals.

  Uninhabited (AreDual Stop (Send msgTy k)) where
    uninhabited DStop impossible

  Uninhabited (AreDual Stop (Recv msgTy k)) where
    uninhabited DStop impossible

  Uninhabited (AreDual (Send x k) (Send y l)) where
    uninhabited DStop impossible

  Uninhabited (AreDual (Recv x k) (Recv y l)) where
    uninhabited DStop impossible

  -- [ NOTE ]
  --
  -- We have only provided *half* the comparisons.
  --
  -- Duality is a symmetric relation between terms. By 'breaking the
  -- symmetry', we can reduce the number of proofs we need to write.

  ||| If `this` is dual to `that` then `that` is dual to `this`.
  swap : AreDual this that -> AreDual that this
  swap DStop = DStop
  swap (DSend prfEQ cont) = DRecv (sym prfEQ) (swap cont)
  swap (DRecv prfEQ cont) = DSend (sym prfEQ) (swap cont)

  ||| If `this` is not dual to `that` then we know that `that` is not dual to `this`.
  negSym : Not (AreDual this that)
        -> AreDual that this
        -> Void
  negSym f x
    = f (swap x)


  ||| Given two sets of actions, is `this` action dual to `that` action.
  areDual : DecEq a -- We need decidable equality of message types!
         => (this : Action a)
         -> (that : Action a)
                 -> Dec (AreDual this that)
                 -- ^ We are decidable as duality *is* decidable.

  areDual (Send payloadType next) (Send x y)
    = No absurd
    --   ^ yes this is absurd

  areDual (Send x k) (Recv y l) with (decEq x y)
    -- First we consider if the message types are equal.
    areDual (Send x k) (Recv x l) | (Yes Refl) with (areDual k l)
      -- Then we check the continuation.
      areDual (Send x k) (Recv x l) | (Yes Refl) | (Yes prf)
        = Yes (DSend Refl prf)
      areDual (Send x k) (Recv x l) | (Yes Refl) | (No contra)
        = No (\case (DSend Refl cont) => contra cont)

    areDual (Send x k) (Recv y l) | (No contra)
      = No (\case (DSend prfEQ cont) => contra prfEQ)

  areDual (Send payloadType next) Stop
    = No (negSym absurd)
    --    ^ reusing an existing proof by swapping terms

  -- Recv is the same as `Send` but with the terms reversed.
  areDual (Recv x k) (Send y l) with (decEq x y)
    areDual (Recv x k) (Send x l) | (Yes Refl) with (areDual k l)
      areDual (Recv x k) (Send x l) | (Yes Refl) | (Yes prf)
        = Yes (DRecv Refl prf)
      areDual (Recv x k) (Send x l) | (Yes Refl) | (No contra)
        = No (\case (DRecv Refl cont) => contra cont)
    areDual (Recv x k) (Send y l) | (No contra)
      = No (\case (DRecv prfEQ cont) => contra prfEQ)
  areDual (Recv x this) (Recv payloadType next)
    = No absurd

  areDual (Recv x this) Stop
    = No (negSym absurd)

  -- `Stop` is straightforward
  areDual Stop (Send payloadType next)
    = No absurd
  areDual Stop (Recv payloadType next)
    = No absurd

  areDual Stop Stop
    = Yes DStop


  ||| Here is another extrinsic proof that our `Dual` function is
  ||| correct.
  |||
  ||| Note the use of `Dual` in the return argument.
  |||
  Dual' : (this : Action a)
               -> AreDual this (Dual this)

  Dual' (Send payloadType next) with (Dual' next)
    Dual' (Send payloadType next) | res = DSend Refl res

  Dual' (Recv payloadType next) with (Dual' next)
    Dual' (Recv payloadType next) | res = DRecv Refl res
  Dual' Stop = DStop

  ||| Here is a way to compute the dual of an action together wit
  ||| proof that the result *is* the dual.
  dual : (this : Action a)
             -> (that ** AreDual this that)

  dual (Send payloadType next) with (dual next)
    dual (Send payloadType next) | (fst ** snd)
      = (Recv payloadType fst ** DSend Refl snd)

  dual (Recv payloadType next) with (dual next)
    dual (Recv payloadType next) | (fst ** snd)
      = (Send payloadType fst ** DRecv Refl snd)

  dual Stop
    = (Stop ** DStop)


  ||| Here we will use the Universe Pattern to resolve message types
  ||| to concret types.
  |||
  ||| *Note* we are resolving down to the simpler definition of `Actions`.
  Resolve : (f : a -> Type)
         -> Action a
         -> Action
  Resolve f (Send ty next)
    = Send (f ty) (Resolve f next)

  Resolve f (Recv ty next)
    = Recv (f ty) (Resolve f next)

  Resolve f Stop
    = Stop


  ||| Here is different type signature of `Resolve` that reuses the
  ||| definition of `Action` from this namespace..
  Resolve' : (f : a -> Type) -> Action a -> Action Type

  -- [ Question ]
  --
  -- Complete the definition.

  -- [ Question ]
  --
  -- Try and formalise `Resolve` as we have done with Dual.

  -- [ Question ]
  --
  -- Action is defined as an indexed type. Can you provide an instance
  -- of Functor for it?

  -- =============================================================================
  --
  ||| Example of Universe Pattern in action.
  --
  -- =============================================================================
  --
  namespace Example

    data Ty = PING | PONG


    EchoClient : Action Ty
    EchoClient
      = Send PING
      $ Recv PONG
        Stop

    EchoServer : Action Ty
    EchoServer
      = Dual (EchoClient)


    InterpTy : Ty -> Type
    InterpTy PING = String
    InterpTy PONG = String

    public export
    EchoClientReal : Simple.Action
    EchoClientReal
      = Resolve InterpTy EchoClient

    public export
    EchoClientServer : Simple.Action
      = Resolve InterpTy EchoServer

-- [ EOF ]


-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--

{- [ Marker ] ==================================================================

== [ Marker ] =============================================================== -}
