--- =============================================================================
---
||| Using Binary Session Types to Type Sockets
---
--- =============================================================================
|||
||| We build upon the continuation style linear and TYPESTATE sockets
||| API, so that errors are handled.
|||
||| For brevity we only look at the operations from a Client's perspective.
|||
||| The exercises will ask you to complete the server variant.
|||
||| The `network` library, packaged as part Idris' standard libraries,
||| provides a working Linear Raw network implementation that you can play
||| with, not to mention a linear one.
|||
|||
module SocketsTyped

namespace Sessions

  public export
  data Action : (payloadTypes : Type) -> Type where

    ||| Send a message of type `payloadType`, and then do the 'next'
    ||| action.
    |||
    Send : (payloadType : a)
        -> (next        : Action a)
                       -> Action a

    ||| Receive a message of type `payloadType`, and then do the 'next'
    ||| action.
    |||
    Recv : (payloadType : a)
        -> (next        : Action a)
                       -> Action a

    ||| Terminate the interaction
    Stop : Action a

  public export
  Dual : Action a -> Action a
  Dual (Send x y) = Recv x (Dual y)
  Dual (Recv x y) = Send x (Dual y)
  Dual Stop = Stop

  public export
  Resolve : (f : a -> Type)
         -> Action a
         -> Action Type
  Resolve f (Send ty next)
    = Send (f ty) (Resolve f next)

  Resolve f (Recv ty next)
    = Recv (f ty) (Resolve f next)

  Resolve f Stop
    = Stop

-- =============================================================================
--
-- Useful datatypes
--
-- =============================================================================
--
-- Rather than define a complete set of states for both server and client sockets,
-- we are concentrating on the client operations.
--
-- As with all the code, We have left some definitions abstract on purpose.

data SState = READY (Action Type)
            | OPEN  (Action Type)
            | CLOSED

data Socket : SState -> Type where

export
data Error : Type where

-- =============================================================================
--
-- State Changing
--
-- =============================================================================
--
-- We need to map actions to states, and determine the `NextSteps` in
-- the finite state machine.
--

||| `ClientAction` maps the expected state of the socket, with an
||| action.
|||
||| We do this when calculating the `NextStep` to ensure that the
||| given state is of the correct 'shape'.
|||
data ClientAction : SState -> Type where
  Connect : ClientAction (READY act)
  Send    : ClientAction (OPEN (Send ty next))
  Receive : ClientAction (OPEN (Recv ty next))
  Close   : ClientAction st

||| We need to track if the output of an error represents a positive or negative result.
|||
||| When using `Maybe`, `Nothing` can sometimes represent a good thing!
data Polarity = POS | NEG

fromMaybe : Maybe a -> Polarity
fromMaybe Nothing = NEG
fromMaybe (Just x) = POS

fromMaybe' : Maybe a -> Polarity
fromMaybe' Nothing = POS
fromMaybe' (Just x) = NEG

fromEither : Either a b -> Polarity
fromEither (Left x) = NEG
fromEither (Right x) = POS

||| `NextStep` calculates the next type of the socket, based on a
||| given polarity, a given socket state and an action.
|||
||| Positive outputs perform the state change to the next positive step.
||| Negative outputs (i.e. errors) changes the socket to closed.
|||
||| We calculate a type rather than a `SState` instance, because of
||| where/how we perform the calculation.
|||
||| @st  The given state of a socket.
||| @act The action that causes the next step.
||| @pol The polarity of a calculated result.
|||
NextStep : (st  : SState)
        -> (act : ClientAction st)
        -> (pol : Polarity)
               -> Type

-- Connecting a ready socket activates the protocol.
NextStep (READY acts) Connect POS
  = Socket (OPEN acts)

-- A successful send shifts the socket to the next step.
NextStep (OPEN (Send ty next)) Send POS
  = Socket (OPEN next)

-- A successful receive shifts the socket to the next step.
NextStep (OPEN (Recv ty next)) Receive POS
  = Socket (OPEN next)

-- Anything else closes the socket.
NextStep _ _  _
  = Socket CLOSED

-- =============================================================================
--
-- The API itself
--
-- =============================================================================
--
-- We now define our API that abstracts over a lower-level API taken
-- from a systems-level library.
--

||| Creates a 'socket' that will follow the given local type.
|||
socket : (act : Action Type)
      -> (cont : (1 sock : Socket (READY act)) -> IO ())
              -> IO ()

||| Tears it all down
close : (1 sock : Socket st) -> IO ()

-- =============================================================================
--
-- Creating Connections
--
-- =============================================================================

||| Connects to a given address and port.
|||
||| Returns Nothing on success, or Just an error.
|||
export
connect : (1 sock : Socket (READY act))
       -> (  addr : String)
       -> (  port : Nat )
       -> (  cont : (1 res : Res (Maybe Error)
                                 (\res => NextStep (READY act) Connect (fromMaybe' res)))
                          -> IO ())
                 -> IO ()

||| Send the string message.
|||
||| Returns on failure a `SocketError`.
||| Returns on success the `ResultCode`.
|||
export
send : (1 sock : Socket (OPEN (Send ty next)))
    -> (  msg  : ty)
    -> (  cont : (1 res : Res (Maybe Error)
                              (\res => NextStep (OPEN (Send ty next))
                                                Send
                                                (fromMaybe' res)))
                       -> IO ())
              -> IO ()

-- [ NOTE ]
--
-- Remember our definition of `Send` has one continuation.
--
-- Select offers multiple choices to select from.

-- [ QUESTION ]
--
-- Can you think how we can modify `send` to select the correct
-- continuation if we shift `Action` to use select and offer?

||| Receive data on the specified socket.
|||
||| Returns on failure a `SocketError`
||| Returns on success a pairing of:
|||
||| + `String`     :: The payload.
||| + `ResultCode` :: The result of the underlying function.
|||
export
recv : (1 sock : Socket (OPEN (Recv ty next)))
    -> (  cont : (1 res : Res (Either Error (ty, Int))
                              (\res => NextStep (OPEN (Recv ty next))
                                                Receive
                                                (fromEither res)))
                           -> IO ())
              -> IO ()

-- [ NOTE ]
--
-- Remember our definition of `Recv` has one continuation.
--
-- Offer offers multiple choices to select from.

-- [ QUESTION ]
--
-- Can you think how we can modify `receive` to offer choices based on
-- the data received?

||| Tears it all down
export
stop : (1 sock : Socket (OPEN Stop)) -> IO ()
stop = close

-- =============================================================================
--
-- Our Echo Protocol(s)
--
-- =============================================================================
--

data Ty = PING | PONG

EchoClient : Action Ty
EchoClient
  = Send PING
  $ Recv PONG
    Stop

EchoServer : Action Ty
EchoServer
  = Dual (EchoClient)


InterpTy : Ty -> Type
InterpTy PING = String
InterpTy PONG = String

public export
EchoClientReal : Action Type
EchoClientReal
  = Resolve InterpTy EchoClient

public export
EchoClientServer : Action Type
EchoClientServer
  = Resolve InterpTy EchoServer

-- =============================================================================
--
-- A Bad Program
--
-- =============================================================================
--

-- [ Question ]
--
-- Can you rewrite the program so that it works correctly.

prog : IO ()
prog
  = socket
  $ \sock =>
  connect sock "localhost" 862
  $ \res =>
    case res of
      ((Just x) # r) => close r
      (Nothing # r)
        => recv r
        $ \res =>
           case res of
             ((Left y) # x) => close x
             ((Right (str,l)) # x)
               => close x

-- [ EOF ]


-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--

{- [ Marker ] ==================================================================

== [ Marker ] =============================================================== -}
