--- =============================================================================
---
||| Mechanising A Simple Expression Language.
---
--- =============================================================================
|||
||| Recall that dependent types provide an expressive environment that
||| enables mechanised reasoning about our programs. It is this very
||| same environment that enables mechanised reasoning about our programming
||| languages!
|||
||| This file demonstrates the power of dependent types to not only
||| enable mechanised reasoning about our programming languages, but
||| how we can repurpose the mechanised proofs as part of a language
||| runtime. That is, the mechanics presented in this file represent
||| the building blocks that lead to the core workings of a verified
||| compiler.
|||
||| Specifically, this file demonstrates:
|||
||| 1. provision of intrinsically-typed (and well-scoped)
|||    representations of our language's abstract syntax;
|||
||| 2. use of predicates to distinguish between terms based on if they
|||    are reducible or not;
|||
||| 3. operations over intrinsically-typed terms to enable
|||    substitution of terms, required for reducing bound variables;
|||
||| 4. using dependent types to encode type-preserving small-step
|||    operational semantics, and encode what it means for progress in
|||    reduction to occur;
|||
||| 5. using the description of progress to present a proof that
|||    well-typed terms will eventually reduce (evaluate) to a
|||    value---our proof of progress;
|||
||| 6. repurposing the 'proof of progress' to evaluate
|||    intrinsically-typed terms to values;
|||
||| The contents on this file are based on Chapter 8 of PLFA.
|||
||| The exercises will ask you to extend the language so that you can
||| add new terms.
|||
module Razor

import Data.List.Elem
import Data.Fuel

%default total

-- =============================================================================
--
||| Intrinsically-Typed Terms: Types, Terms, and Values!
--
-- =============================================================================
|||
||| We present the intrinsically-typed terms representing our
||| language's abstract syntax.
namespace Terms

  public export
  data Ty = BOOL | NAT

  public export
  data Term : (  context : List Ty)
           -> (0 type    : Ty)
                      -> Type
    where
      -- Constants
      B : Bool -> Term ctxt BOOL
      N : Nat  -> Term ctxt NAT

      -- [ NOTE ]
      --
      -- For practical reasons, we have encoded our constants using
      -- Idris primitives. We could have instead provided a synthetic
      -- representation that uses data constructors to encode the
      -- constants, but this representation is a little easier.

      -- Operations

      Add : (l,r : Term ctxt NAT)
                -> Term ctxt NAT

      And : (l,r : Term ctxt BOOL)
                -> Term ctxt BOOL

      -- [ QUESTION ]
      --
      -- Can you think how dependent types enable us to provide a
      -- single data structure for operations?

      -- Binding

      Var : (idx : Elem type ctxt)
                -> Term ctxt type

      Let : {typeT, typeB : Ty}
         -> (this : Term ctxt typeT)
         -> (body : Term (typeT :: ctxt) typeB)
                 -> Term ctxt typeB

      -- [ NOTE ]
      --
      -- Variables are 'De Bruijn' encoded to provide a nameless
      -- representation. We abandon the use of string names in favour
      -- of representing *when* a bound variable is introduced.
      --
      -- We have used a 'normal list' representation such that our
      -- lists 'grow to the left'.  A problem with this approach is
      -- that the head of the context is the *latest* bound
      -- variable. There is a growing movement to use a 'SnocList'
      -- (lists that 'grow to the right') as they better capture how
      -- contexts grow such that the head of the context is the
      -- *first* bound variable.

  ||| Some examples of well-typed and ill-typed terms.
  namespace Example
    public export
    example : Term Nil BOOL
    example
      = And (B True) (B False)

    --export
    --example1 : Term Nil NAT
    --example1
    --  = Let example
    --        (Add (N 1) (Var Here))

    public export
    example2 : Term Nil BOOL
    example2
      = Let example
            (And (B False) (Var Here))

  ||| Here we use a predicate to state that our language's values are
  ||| 'just' our language's constants.
  namespace Value

    public export
    data Value : (term : Term ctxt type)
                      -> Type
      where
        N : Value (N n)
        B : Value (B b)



-- =============================================================================
--
||| Renaming Bound terms.
--
-- =============================================================================
|||
||| With an intrinsically-typed representation, we have removed
||| concrete names from our language's abstract syntax. To enable
||| substitution of variables with terms, we need to capture the
||| renaming of nameless variables as the typing context is
||| manipulated.
|||
||| Renaming, and as we will see later on, and substitution are
||| traversals over our terms, such that the context is correctly
||| updated and our term's types are preserved.
namespace Renaming

  ||| Updates a variable's typing context with a new context in which
  ||| a new variable has been introduced.
  |||
  ||| Takes a function that updates nameless variable from one context
  ||| to another, and uses that function to update a function that
  ||| updates a variable in which the context has grown by a new
  ||| variable.
  |||
  public export
  weaken : (f : Elem type           old  -> Elem type           new )
        -> (    Elem type (type' :: old) -> Elem type (type' :: new))

  weaken f Here
    = Here
  weaken f (There rest)
    = There (f rest)

  ||| Given a function that updates variables from one context to
  ||| another, `rename` will traverse the program to update variables
  ||| with the new context.
  public export
  rename : (f : forall ty . Elem ty old    -> Elem ty new)
        -> (    forall ty . Term    old ty -> Term    new ty)

  rename f (Var x)
    = Var (f x)
    --    ^ Where the variable is renamed. The rest is traversing
    --      terms to update all variables.
  rename f (B x)
    = B x

  rename f (N k)
    = N k

  rename f (Add l r)
    = Add (rename f l) (rename f r)

  rename f (And l r)
    = And (rename f l) (rename f r)

  rename f (Let this body)
    = Let (rename f this) (rename (weaken f) body)

-- =============================================================================
--
||| Substitution Variables with Terms
--
-- =============================================================================
|||
||| With an intrinsically-typed representation, we also need to reduce
||| bound variables. We do so by using renaming to update bound
||| variables when the number of terms have been reduced.
||| Substitution works similarly.
|||
namespace Substitution

  ||| Updates variables and terms with a new typing context.
  public export
  weakens : {old, new : List Ty}

         -> (f : {ty : Ty}
              -> Elem ty old -> Term new ty)

         -> ({ty,type' : Ty}
              -> Elem ty  (type' :: old)
              -> Term     (type' :: new) ty)

  weakens f Here
     = Var Here
  weakens f (There rest)
     = rename There (f rest)


  ||| General substitution where we substitute the bound variable with
  ||| a term, and update the typing context accordingly.
  public export
  subst : {old, new : List Ty}
       -> (f : {ty : Ty}
             -> Elem ty old
             -> Term    new ty)
       -> ({ty : Ty}
              -> Term old ty
              -> Term new ty)

  subst f (Var x)
    = f x
    -- ^ Here is where the substitution occurs. The rest is traversing
    --   terms to update all occurrences of the variable.

  subst f (B x)
    = B x

  subst f (N k)
    = N k

  subst f (Add l r)
    = Add (subst f l) (subst f r)

  subst f (And l r)
    = And (subst f l) (subst f r)

  subst f (Let this body)
    = Let (subst f this) (subst (weakens f) body)

  ||| Substituting a Single term.
  |||
  namespace Single

    ||| Substitutes the term `this` in the scope `inThis`.
    |||
    ||| Requires `this` to at the head of the context i.e. the most
    ||| recently introduced variable.
    public export
    subst : {typeA,typeB : Ty}
         -> {ctxt : List Ty}
         -> (this          : Term           ctxt  typeB)
         -> (inThis        : Term (typeB :: ctxt) typeA)
                          -> Term           ctxt  typeA

    -- [ NOTE]
    --
    -- The context of `this` and the resulting term, as compared to
    -- `inThis` *and* the definition of `Let`.

    subst {ctxt} {typeA} {typeB} this inThis
        = subst (apply this) inThis

      where

        ||| Apply is the function that swaps out the variable for a
        ||| term.
        apply : {typeA,typeB : Ty}
             -> {ctxt        : List Ty}
             -> (this   : Term                 ctxt   typeB)
             -> (idx    : Elem typeA (typeB :: ctxt))
                       -> Term                 ctxt   typeA
        apply this Here
          = this

        apply this (There rest)
          = Var rest

-- =============================================================================
--
||| Encoding Operational Semantics
--
-- =============================================================================
|||
||| Here we will use a dependent type to encode our language's
||| operational semantics.
|||
||| *Note* that our reductions will be type-preserving by design!.
|||
namespace Progress

  ||| Small-step operational semantics for our language.
  |||
  ||| We get preservation for 'free' by ensuring that the input and
  ||| output terms have the same typing context and type.
  |||
  public export
  data Reduce : (this : Term ctxt type)
             -> (that : Term ctxt type)
             --              ^ The same Context!
             --                   ^ The same type!
                     -> Type
    where
      -- Operations to Reduce Add:

      ||| We begin by reducing the left operand to a value.
      SimplifyAddL : Reduce      this that
                  -> Reduce (Add this      right)
                            (Add      that right)

      ||| We only reduce the right operand, if the left is a value.
      SimplifyAddR : Value       left
                  -> Reduce           this that
                  -> Reduce (Add left this     )
                            (Add left      that)

      ||| When both operands are values, we can reduce the value.
      ReduceAdd : Reduce (Add (N a) (N b))
                         (N     (a +   b))

      -- Operations to Reduce And:

      ||| We begin by reducing the left operand to a value.
      SimplifyAndL : this `Reduce` that
                  --------------------------------
                  -> Reduce (And this      right)
                            (And      that right)

      ||| We only reduce the right operand, if the left is a value.
      SimplifyAndR : Value       left
                  -> Reduce           this that
                  -> Reduce (And left this     )
                            (And left      that)

      ||| When both operands are values, we can reduce the value.
      ReduceAnd : Reduce (And (B a)    (B b))
                         (B   (  a  &&    b))

      -- Binders

      ||| Reduce the bound term first.
      SimplifyLet : Reduce      this that
                 -> Reduce (Let this      body)
                           (Let      that body)

      ||| Once we have a boudn *value* then substitute it in.
      ReduceLet : Value                this
               -> Reduce (Let          this body)
                         (Single.subst this body)

  -- [ QUESTION ]
  --
  -- Our operational semantics do not short circuit `Add` and `And`
  -- during reduction *if* `Zero` or `False` are encountered.  Should we
  -- do so?

  -- [ QUESTION ]
  --
  -- `Reduce` is a simple transformation of terms. Can you think of how
  -- you can use dependent types to write syntax transformations that
  -- are type safe?  For example, mapping ternary (or n-ary) operations
  -- into binary ones?


  ||| Encodes that reducing closed terms, terms in which there are no
  ||| bound variables, by a single step will either produce a value,
  ||| or require more reductions to be performed.
  public export
  data Progress : (term : Term Nil type)
                       -> Type
    where
      ||| The term we wish to reduce *is* a value.
      Stop : (value : Value    term)
                   -> Progress term

      ||| We can perform a reduction of `this` terms to `that` term.
      |||
      ||| `that` has been qualified as we will need to access it later
      ||| on.
      Step : {that : Term Nil type}
          -> (step : Reduce   this that)
                  -> Progress this

  ||| Single-step Progress demonstrates that we can reduce a closed
  ||| term, one step at a time.
  export
  progress : (term : Term Nil type)
                  -> Progress term

  -- [ NOTE ]
  --
  -- Impossible in a close term!
  progress (Var _) impossible

  -- Constants *are* values!
  progress (B x) = Stop B
  progress (N k) = Stop N

  -- Operations

  progress (Add l r) with (progress l)
    progress (Add (N n) r) | (Stop N) with (progress r)
      progress (Add (N n) (N m)) | (Stop N) | (Stop N)
        = Step ReduceAdd
      progress (Add (N n) r) | (Stop N) | (Step step)
        = Step (SimplifyAddR N step)
    progress (Add l r) | (Step step)
      = Step (SimplifyAddL step)

  progress (And l r) with (progress l)
    progress (And (B n) r) | (Stop B) with (progress r)
      progress (And (B n) (B m)) | (Stop B) | (Stop B)
        = Step ReduceAnd
      progress (And (B n) r) | (Stop B) | (Step step)
        = Step (SimplifyAndR B step)
    progress (And l r) | (Step step)
      = Step (SimplifyAndL step)

  -- Let binders

  progress (Let this body) with (progress this)
    progress (Let this body) | (Stop value)
      = Step (ReduceLet value)
    progress (Let this body) | (Step step)
      = Step (SimplifyLet step)

-- =============================================================================
--
||| Evaluation, chains steps together!
--
-- =============================================================================
|||
||| Our `Progress` namespace highlights how to do one reduction step
||| at a time. Given terms in our language, we need to potentially
||| chain these steps together to produce a value.
|||
||| It is important to not that evaluation may take infinitely many
||| steps. We use the `Inf` construct to capture potentially infinite
||| computation.  However, as Idris is a total language it enables
||| reasoning about program termination, and thus evaluation will not
||| be seen as 'total' because of `Inf`.
|||
||| To reason about infinite computation, we use the `Fuel` data
||| structure to capture that there may be an infinite supply of fuel
||| that may become exhausted.  When evaluating terms we use
||| coinduction (of Inf and Fuel) to capture that the evaluation may
||| eventually terminate, thus satisfying totality checker.
namespace Evaluation

  ||| `Reduces` is a list of reduction steps that take `this` term to
  ||| an irreducible version `that`.
  |||
  public export
  data Reduces : (this,that : Term ctxt type)
                           -> Type
    where

      ||| A value will produce the 'empty' list, i.e. no steps are
      ||| required.
      Nil : Reduces this this

      ||| A reducible value will require a least one step of reduction
      ||| to produce a value.
      (::) : Reduce  this that
          -> Reduces      that end
          -> Reduces this      end


  ||| We know that evaluation of terms finishes *when* the term is a
  ||| value or there is no more fuel.
  public export
  data Finished : (term : Term ctxt type)
                       -> Type
    where
      ||| We can stop.
      IsValue : Value    term
             -> Finished term

      ||| We have ran out of fuel.
      NoFuel : Finished term

  ||| Evaluation reduces a term down to a finished 'state'.
  |||
  ||| The reduction may take an infinite amount of steps, and may exhauste our fuel.
  public export
  data Evaluation : (term : Term Nil type)
                         -> Type
    where
      Eval : {this, that : Term Nil type}
          -> (steps  : Inf (Reduces this that))
          -> (result : Finished          that)
                    -> Evaluation   this

  ||| Evaluates a closed term until we return a value or have ran out of fuel.
  |||
  ||| We use our operational semantics to perform the reduction, and
  ||| carry on until there are more reductions to be made (or ran out
  ||| of fuel).
  |||
  public export
  evaluate : (fuel : Fuel)
          -> (term : Term Nil type)
                  -> Evaluation term

  -- No more fuel.
  evaluate Dry term
    = Eval Nil NoFuel

  -- if there is fuel, then perform a reduction step.
  evaluate (More fuel) term with (progress term)
    -- if we reduce to a value then stop
    evaluate (More fuel) term | (Stop value)
      = Eval Nil (IsValue value)

    -- otherwise keep on reducing on the resulting step.
    --
    -- *Note* how we use the implicitly bound type-level `that` for the next step!
    -- `that` has been calculated
    evaluate (More fuel) term | (Step step {that}) with (evaluate fuel that)
      evaluate (More fuel) term | (Step step {that = that}) | (Eval steps result)
        = Eval (step :: steps) result

-- =============================================================================
--
||| The Fun Part! Running programs.
--
-- =============================================================================
|||
namespace Run

  ||| Evalute a term and return the result of evaluation, together
  ||| with the evaluation steps performed.
  |||
  ||| Notice that we export the function as *covering*, `forever` is
  ||| not total, but `evaluate` is!  `run` is the only part of our
  ||| core that is not total i.e. not guaranteed to terminate.
  |||
  ||| Also notice that we return `Maybe`, we need to encapsulate that
  ||| our computation *may* have (albeit unlikely) have run out of
  ||| fuel for computation.
  export covering
  run : {0 type : Ty}
     -> (this : Term Nil type)
             -> Maybe (that : Term Nil type ** Reduces this that)
  run this with (evaluate forever this)
    run this | (Eval steps (IsValue {term} x))
      = Just (term ** steps)
    run this | (Eval steps NoFuel)
      = Nothing

namespace Terms
  namespace Example
    public export
    example3 : Term Nil NAT
    example3
      = Let (Let (N 1) (Add (N 3) (Add (N 5) (Var Here))))
            (Let (B False)
                 (Add (Var (There Here)) (Let (N 3) (Var Here))))

{- [ Marker ] ==================================================================
== [ Marker ] =============================================================== -}

-- [ EOF ]

-- [ Question ]
--
--

-- [ NOTES ]
--
--

-- =============================================================================
--
-- =============================================================================
--
--

{- [ Marker ] ==================================================================

== [ Marker ] =============================================================== -}
